# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Correctness test for tf.keras Embedding models using DistributionStrategy."""

import numpy as np
import tensorflow.compat.v2 as tf

import keras
from keras.distribute import keras_correctness_test_base
from keras.optimizers.legacy import gradient_descent as gradient_descent_keras


class DistributionStrategyEmbeddingModelCorrectnessTest(
    keras_correctness_test_base.TestDistributionStrategyEmbeddingModelCorrectnessBase  # noqa: E501
):
    def get_model(
        self,
        max_words=10,
        initial_weights=None,
        distribution=None,
        input_shapes=None,
    ):
        del input_shapes
        with keras_correctness_test_base.MaybeDistributionScope(distribution):
            word_ids = keras.layers.Input(
                shape=(max_words,), dtype=np.int32, name="words"
            )
            word_embed = keras.layers.Embedding(input_dim=20, output_dim=10)(
                word_ids
            )
            if self.use_distributed_dense:
                word_embed = keras.layers.TimeDistributed(
                    keras.layers.Dense(4)
                )(word_embed)
            avg = keras.layers.GlobalAveragePooling1D()(word_embed)
            preds = keras.layers.Dense(2, activation="softmax")(avg)
            model = keras.Model(inputs=[word_ids], outputs=[preds])

            if initial_weights:
                model.set_weights(initial_weights)

            model.compile(
                optimizer=gradient_descent_keras.SGD(learning_rate=0.1),
                loss="sparse_categorical_crossentropy",
                metrics=["sparse_categorical_accuracy"],
            )
        return model

    @tf.__internal__.distribute.combinations.generate(
        keras_correctness_test_base.test_combinations_for_embedding_model()
        + keras_correctness_test_base.multi_worker_mirrored_eager()
    )
    def test_embedding_model_correctness(
        self, distribution, use_numpy, use_validation_data
    ):

        self.use_distributed_dense = False
        self.run_correctness_test(distribution, use_numpy, use_validation_data)

    @tf.__internal__.distribute.combinations.generate(
        keras_correctness_test_base.test_combinations_for_embedding_model()
        + keras_correctness_test_base.multi_worker_mirrored_eager()
    )
    def test_embedding_time_distributed_model_correctness(
        self, distribution, use_numpy, use_validation_data
    ):
        self.use_distributed_dense = True
        self.run_correctness_test(distribution, use_numpy, use_validation_data)


class DistributionStrategySiameseEmbeddingModelCorrectnessTest(
    keras_correctness_test_base.TestDistributionStrategyEmbeddingModelCorrectnessBase  # noqa: E501
):
    def get_model(
        self,
        max_words=10,
        initial_weights=None,
        distribution=None,
        input_shapes=None,
    ):
        del input_shapes
        with keras_correctness_test_base.MaybeDistributionScope(distribution):
            word_ids_a = keras.layers.Input(
                shape=(max_words,), dtype=np.int32, name="words_a"
            )
            word_ids_b = keras.layers.Input(
                shape=(max_words,), dtype=np.int32, name="words_b"
            )

            def submodel(embedding, word_ids):
                word_embed = embedding(word_ids)
                rep = keras.layers.GlobalAveragePooling1D()(word_embed)
                return keras.Model(inputs=[word_ids], outputs=[rep])

            word_embed = keras.layers.Embedding(
                input_dim=20,
                output_dim=10,
                input_length=max_words,
                embeddings_initializer=keras.initializers.RandomUniform(0, 1),
            )

            a_rep = submodel(word_embed, word_ids_a).outputs[0]
            b_rep = submodel(word_embed, word_ids_b).outputs[0]
            sim = keras.layers.Dot(axes=1, normalize=True)([a_rep, b_rep])

            model = keras.Model(inputs=[word_ids_a, word_ids_b], outputs=[sim])

            if initial_weights:
                model.set_weights(initial_weights)

            # TODO(b/130808953): Switch back to the V1 optimizer after
            # global_step is made mirrored.
            model.compile(
                optimizer=gradient_descent_keras.SGD(learning_rate=0.1),
                loss="mse",
                metrics=["mse"],
            )
        return model

    def get_data(
        self,
        count=(
            keras_correctness_test_base._GLOBAL_BATCH_SIZE
            * keras_correctness_test_base._EVAL_STEPS
        ),
        min_words=5,
        max_words=10,
        max_word_id=19,
        num_classes=2,
    ):
        features_a, labels_a, _ = super().get_data(
            count, min_words, max_words, max_word_id, num_classes
        )

        features_b, labels_b, _ = super().get_data(
            count, min_words, max_words, max_word_id, num_classes
        )

        y_train = np.zeros((count, 1), dtype=np.float32)
        y_train[labels_a == labels_b] = 1.0
        y_train[labels_a != labels_b] = -1.0
        # TODO(b/123360757): Add tests for using list as inputs for multi-input
        # models.
        x_train = {
            "words_a": features_a,
            "words_b": features_b,
        }
        x_predict = x_train

        return x_train, y_train, x_predict

    @tf.__internal__.distribute.combinations.generate(
        keras_correctness_test_base.test_combinations_for_embedding_model()
        + keras_correctness_test_base.multi_worker_mirrored_eager()
    )
    def test_siamese_embedding_model_correctness(
        self, distribution, use_numpy, use_validation_data
    ):
        self.run_correctness_test(distribution, use_numpy, use_validation_data)


if __name__ == "__main__":
    tf.__internal__.distribute.multi_process_runner.test_main()



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Keras models API.
"""

import sys as _sys

from keras.engine.sequential import Sequential
from keras.engine.training import Model
from keras.models.cloning import clone_model
from keras.premade_models.linear import LinearModel
from keras.premade_models.wide_deep import WideDeepModel
from keras.saving.legacy.model_config import model_from_config
from keras.saving.legacy.model_config import model_from_json
from keras.saving.legacy.model_config import model_from_yaml
from keras.saving.saving_api import load_model
from keras.saving.saving_api import save_model
from tensorflow.python.util import module_wrapper as _module_wrapper

if not isinstance(_sys.modules[__name__], _module_wrapper.TFModuleWrapper):
  _sys.modules[__name__] = _module_wrapper.TFModuleWrapper(
      _sys.modules[__name__], "keras.models", public_apis=None, deprecation=True,
      has_lite=False)



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Layer that computes the dot product between two inputs."""


import tensorflow.compat.v2 as tf

from keras import backend
from keras.engine import base_layer_utils
from keras.layers.merging.base_merge import _Merge
from keras.utils import tf_utils

# isort: off
from tensorflow.python.util.tf_export import keras_export


@keras_export("keras.layers.Dot")
class Dot(_Merge):
    """Layer that computes a dot product between samples in two tensors.

    E.g. if applied to a list of two tensors `a` and `b` of shape
    `(batch_size, n)`, the output will be a tensor of shape `(batch_size, 1)`
    where each entry `i` will be the dot product between
    `a[i]` and `b[i]`.

    >>> x = np.arange(10).reshape(1, 5, 2)
    >>> print(x)
    [[[0 1]
      [2 3]
      [4 5]
      [6 7]
      [8 9]]]
    >>> y = np.arange(10, 20).reshape(1, 2, 5)
    >>> print(y)
    [[[10 11 12 13 14]
      [15 16 17 18 19]]]
    >>> tf.keras.layers.Dot(axes=(1, 2))([x, y])
    <tf.Tensor: shape=(1, 2, 2), dtype=int64, numpy=
    array([[[260, 360],
            [320, 445]]])>

    >>> x1 = tf.keras.layers.Dense(8)(np.arange(10).reshape(5, 2))
    >>> x2 = tf.keras.layers.Dense(8)(np.arange(10, 20).reshape(5, 2))
    >>> dotted = tf.keras.layers.Dot(axes=1)([x1, x2])
    >>> dotted.shape
    TensorShape([5, 1])


    """

    def __init__(self, axes, normalize=False, **kwargs):
        """Initializes a layer that computes the element-wise dot product.

          >>> x = np.arange(10).reshape(1, 5, 2)
          >>> print(x)
          [[[0 1]
            [2 3]
            [4 5]
            [6 7]
            [8 9]]]
          >>> y = np.arange(10, 20).reshape(1, 2, 5)
          >>> print(y)
          [[[10 11 12 13 14]
            [15 16 17 18 19]]]
          >>> tf.keras.layers.Dot(axes=(1, 2))([x, y])
          <tf.Tensor: shape=(1, 2, 2), dtype=int64, numpy=
          array([[[260, 360],
                  [320, 445]]])>

        Args:
          axes: Integer or tuple of integers,
            axis or axes along which to take the dot product. If a tuple, should
            be two integers corresponding to the desired axis from the first
            input and the desired axis from the second input, respectively. Note
            that the size of the two selected axes must match.
          normalize: Whether to L2-normalize samples along the
            dot product axis before taking the dot product.
            If set to True, then the output of the dot product
            is the cosine proximity between the two samples.
          **kwargs: Standard layer keyword arguments.
        """
        super().__init__(**kwargs)
        if not isinstance(axes, int):
            if not isinstance(axes, (list, tuple)):
                raise TypeError(
                    "Invalid type for argument `axes`: it should be "
                    f"a list or an int. Received: axes={axes}"
                )
            if len(axes) != 2:
                raise ValueError(
                    "Invalid format for argument `axes`: it should contain two "
                    f"elements. Received: axes={axes}"
                )
            if not isinstance(axes[0], int) or not isinstance(axes[1], int):
                raise ValueError(
                    "Invalid format for argument `axes`: list elements should "
                    f"be integers. Received: axes={axes}"
                )
        self.axes = axes
        self.normalize = normalize
        self.supports_masking = True
        self._reshape_required = False

    @tf_utils.shape_type_conversion
    def build(self, input_shape):
        # Used purely for shape validation.
        if not isinstance(input_shape[0], tuple) or len(input_shape) != 2:
            raise ValueError(
                "A `Dot` layer should be called on a list of 2 inputs. "
                f"Received: input_shape={input_shape}"
            )
        shape1 = input_shape[0]
        shape2 = input_shape[1]
        if shape1 is None or shape2 is None:
            return
        if isinstance(self.axes, int):
            if self.axes < 0:
                axes = [self.axes % len(shape1), self.axes % len(shape2)]
            else:
                axes = [self.axes] * 2
        else:
            axes = self.axes
        if shape1[axes[0]] != shape2[axes[1]]:
            raise ValueError(
                "Incompatible input shapes: "
                f"axis values {shape1[axes[0]]} (at axis {axes[0]}) != "
                f"{shape2[axes[1]]} (at axis {axes[1]}). "
                f"Full input shapes: {shape1}, {shape2}"
            )

    def _merge_function(self, inputs):
        base_layer_utils.no_ragged_support(inputs, self.name)
        if len(inputs) != 2:
            raise ValueError(
                "A `Dot` layer should be called on exactly 2 inputs. "
                f"Received: inputs={inputs}"
            )
        x1 = inputs[0]
        x2 = inputs[1]
        if isinstance(self.axes, int):
            if self.axes < 0:
                axes = [
                    self.axes % backend.ndim(x1),
                    self.axes % backend.ndim(x2),
                ]
            else:
                axes = [self.axes] * 2
        else:
            axes = []
            for i in range(len(self.axes)):
                if self.axes[i] < 0:
                    axes.append(self.axes[i] % backend.ndim(inputs[i]))
                else:
                    axes.append(self.axes[i])
        if self.normalize:
            x1 = tf.linalg.l2_normalize(x1, axis=axes[0])
            x2 = tf.linalg.l2_normalize(x2, axis=axes[1])
        output = backend.batch_dot(x1, x2, axes)
        return output

    @tf_utils.shape_type_conversion
    def compute_output_shape(self, input_shape):
        if not isinstance(input_shape, (tuple, list)) or len(input_shape) != 2:
            raise ValueError(
                "A `Dot` layer should be called on a list of 2 inputs. "
                f"Received: input_shape={input_shape}"
            )
        shape1 = list(input_shape[0])
        shape2 = list(input_shape[1])
        if isinstance(self.axes, int):
            if self.axes < 0:
                axes = [self.axes % len(shape1), self.axes % len(shape2)]
            else:
                axes = [self.axes] * 2
        else:
            axes = self.axes
        shape1.pop(axes[0])
        shape2.pop(axes[1])
        shape2.pop(0)
        output_shape = shape1 + shape2
        if len(output_shape) == 1:
            output_shape += [1]
        return tuple(output_shape)

    def compute_mask(self, inputs, mask=None):
        return None

    def get_config(self):
        config = {
            "axes": self.axes,
            "normalize": self.normalize,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))


@keras_export("keras.layers.dot")
def dot(inputs, axes, normalize=False, **kwargs):
    """Functional interface to the `Dot` layer.

    Args:
        inputs: A list of input tensors (at least 2).
        axes: Integer or tuple of integers,
            axis or axes along which to take the dot product.
        normalize: Whether to L2-normalize samples along the
            dot product axis before taking the dot product.
            If set to True, then the output of the dot product
            is the cosine proximity between the two samples.
        **kwargs: Standard layer keyword arguments.

    Returns:
        A tensor, the dot product of the samples from the inputs.
    """
    return Dot(axes=axes, normalize=normalize, **kwargs)(inputs)



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Classes and functions implementing to Model SavedModel serialization."""

from keras.saving.legacy import saving_utils
from keras.saving.legacy.saved_model import constants
from keras.saving.legacy.saved_model import layer_serialization
from keras.saving.legacy.saved_model import save_impl


class ModelSavedModelSaver(layer_serialization.LayerSavedModelSaver):
    """Model SavedModel serialization."""

    @property
    def object_identifier(self):
        return constants.MODEL_IDENTIFIER

    def _python_properties_internal(self):
        metadata = super()._python_properties_internal()
        # Network stateful property is dependent on the child layers.
        metadata.pop("stateful")
        metadata["is_graph_network"] = self.obj._is_graph_network
        spec = self.obj.save_spec(dynamic_batch=False)
        metadata["full_save_spec"] = spec
        # save_spec is saved for forward compatibility on older TF versions.
        metadata["save_spec"] = None if spec is None else spec[0][0]

        metadata.update(
            saving_utils.model_metadata(
                self.obj, include_optimizer=True, require_config=False
            )
        )
        return metadata

    def _get_serialized_attributes_internal(self, serialization_cache):
        default_signature = None

        # Create a default signature function if this is the only object in the
        # cache (i.e. this is the root level object).
        if len(serialization_cache[constants.KERAS_CACHE_KEY]) == 1:
            default_signature = save_impl.default_save_signature(self.obj)

        # Other than the default signature function, all other attributes match
        # with the ones serialized by Layer.
        objects, functions = super()._get_serialized_attributes_internal(
            serialization_cache
        )
        functions["_default_save_signature"] = default_signature
        return objects, functions


class SequentialSavedModelSaver(ModelSavedModelSaver):
    @property
    def object_identifier(self):
        return constants.SEQUENTIAL_IDENTIFIER



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras base class for convolution layers."""


import tensorflow.compat.v2 as tf

from keras import activations
from keras import constraints
from keras import initializers
from keras import regularizers
from keras.engine.base_layer import Layer
from keras.engine.input_spec import InputSpec
from keras.utils import conv_utils


class Conv(Layer):
    """Abstract N-D convolution layer (private, used as implementation base).

    This layer creates a convolution kernel that is convolved
    (actually cross-correlated) with the layer input to produce a tensor of
    outputs. If `use_bias` is True (and a `bias_initializer` is provided),
    a bias vector is created and added to the outputs. Finally, if
    `activation` is not `None`, it is applied to the outputs as well.

    Note: layer attributes cannot be modified after the layer has been called
    once (except the `trainable` attribute).

    Args:
      rank: An integer, the rank of the convolution, e.g. "2" for 2D
        convolution.
      filters: Integer, the dimensionality of the output space (i.e. the number
        of filters in the convolution). Could be "None", eg in the case of
        depth wise convolution.
      kernel_size: An integer or tuple/list of n integers, specifying the
        length of the convolution window.
      strides: An integer or tuple/list of n integers,
        specifying the stride length of the convolution.
        Specifying any stride value != 1 is incompatible with specifying
        any `dilation_rate` value != 1.
      padding: One of `"valid"`,  `"same"`, or `"causal"` (case-insensitive).
        `"valid"` means no padding. `"same"` results in padding with zeros
        evenly to the left/right or up/down of the input such that output has
        the same height/width dimension as the input. `"causal"` results in
        causal (dilated) convolutions, e.g. `output[t]` does not depend on
        `input[t+1:]`.
      data_format: A string, one of `channels_last` (default) or
        `channels_first`.
        The ordering of the dimensions in the inputs.
        `channels_last` corresponds to inputs with shape
        `(batch_size, ..., channels)` while `channels_first` corresponds to
        inputs with shape `(batch_size, channels, ...)`.
      dilation_rate: An integer or tuple/list of n integers, specifying
        the dilation rate to use for dilated convolution.
        Currently, specifying any `dilation_rate` value != 1 is
        incompatible with specifying any `strides` value != 1.
      groups: A positive integer specifying the number of groups in which the
        input is split along the channel axis. Each group is convolved
        separately with `filters / groups` filters. The output is the
        concatenation of all the `groups` results along the channel axis.
        Input channels and `filters` must both be divisible by `groups`.
      activation: Activation function to use.
        If you don't specify anything, no activation is applied.
      use_bias: Boolean, whether the layer uses a bias.
      kernel_initializer: An initializer for the convolution kernel. If None,
        the default initializer (glorot_uniform) will be used.
      bias_initializer: An initializer for the bias vector. If None, the default
        initializer (zeros) will be used.
      kernel_regularizer: Optional regularizer for the convolution kernel.
      bias_regularizer: Optional regularizer for the bias vector.
      activity_regularizer: Optional regularizer function for the output.
      kernel_constraint: Optional projection function to be applied to the
          kernel after being updated by an `Optimizer` (e.g. used to implement
          norm constraints or value constraints for layer weights). The function
          must take as input the unprojected variable and must return the
          projected variable (which must have the same shape). Constraints are
          not safe to use when doing asynchronous distributed training.
      bias_constraint: Optional projection function to be applied to the
          bias after being updated by an `Optimizer`.
    """

    def __init__(
        self,
        rank,
        filters,
        kernel_size,
        strides=1,
        padding="valid",
        data_format=None,
        dilation_rate=1,
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        trainable=True,
        name=None,
        conv_op=None,
        **kwargs,
    ):
        super().__init__(
            trainable=trainable,
            name=name,
            activity_regularizer=regularizers.get(activity_regularizer),
            **kwargs,
        )
        self.rank = rank

        if isinstance(filters, float):
            filters = int(filters)
        if filters is not None and filters <= 0:
            raise ValueError(
                "Invalid value for argument `filters`. "
                "Expected a strictly positive value. "
                f"Received filters={filters}."
            )
        self.filters = filters
        self.groups = groups or 1
        self.kernel_size = conv_utils.normalize_tuple(
            kernel_size, rank, "kernel_size"
        )
        self.strides = conv_utils.normalize_tuple(
            strides, rank, "strides", allow_zero=True
        )
        self.padding = conv_utils.normalize_padding(padding)
        self.data_format = conv_utils.normalize_data_format(data_format)
        self.dilation_rate = conv_utils.normalize_tuple(
            dilation_rate, rank, "dilation_rate"
        )

        self.activation = activations.get(activation)
        self.use_bias = use_bias

        self.kernel_initializer = initializers.get(kernel_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)
        self.kernel_constraint = constraints.get(kernel_constraint)
        self.bias_constraint = constraints.get(bias_constraint)
        self.input_spec = InputSpec(min_ndim=self.rank + 2)

        self._validate_init()
        self._is_causal = self.padding == "causal"
        self._channels_first = self.data_format == "channels_first"
        self._tf_data_format = conv_utils.convert_data_format(
            self.data_format, self.rank + 2
        )

    def _validate_init(self):
        if self.filters is not None and self.filters % self.groups != 0:
            raise ValueError(
                "The number of filters must be evenly divisible by the "
                "number of groups. Received: groups={}, filters={}".format(
                    self.groups, self.filters
                )
            )

        if not all(self.kernel_size):
            raise ValueError(
                "The argument `kernel_size` cannot contain 0(s). Received: %s"
                % (self.kernel_size,)
            )

        if not all(self.strides):
            raise ValueError(
                "The argument `strides` cannot contains 0(s). Received: %s"
                % (self.strides,)
            )

        if self.padding == "causal":

            from keras.layers.convolutional.conv1d import Conv1D
            from keras.layers.convolutional.separable_conv1d import (
                SeparableConv1D,
            )

            if not isinstance(self, (Conv1D, SeparableConv1D)):
                raise ValueError(
                    "Causal padding is only supported for `Conv1D`"
                    "and `SeparableConv1D`."
                )

        if max(self.strides) > 1 and max(self.dilation_rate) > 1:
            raise ValueError(
                "`strides > 1` not supported in conjunction with "
                f"`dilation_rate > 1`. Received: strides={self.strides} and "
                f"dilation_rate={self.dilation_rate}"
            )

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape)
        input_channel = self._get_input_channel(input_shape)
        if input_channel % self.groups != 0:
            raise ValueError(
                "The number of input channels must be evenly divisible by "
                "the number of groups. Received groups={}, but the input "
                "has {} channels (full input shape is {}).".format(
                    self.groups, input_channel, input_shape
                )
            )
        kernel_shape = self.kernel_size + (
            input_channel // self.groups,
            self.filters,
        )

        # compute_output_shape contains some validation logic for the input
        # shape, and make sure the output shape has all positive dimensions.
        self.compute_output_shape(input_shape)

        self.kernel = self.add_weight(
            name="kernel",
            shape=kernel_shape,
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
            trainable=True,
            dtype=self.dtype,
        )
        if self.use_bias:
            self.bias = self.add_weight(
                name="bias",
                shape=(self.filters,),
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                trainable=True,
                dtype=self.dtype,
            )
        else:
            self.bias = None
        channel_axis = self._get_channel_axis()
        self.input_spec = InputSpec(
            min_ndim=self.rank + 2, axes={channel_axis: input_channel}
        )
        self.built = True

    def convolution_op(self, inputs, kernel):
        if self.padding == "causal":
            tf_padding = "VALID"  # Causal padding handled in `call`.
        elif isinstance(self.padding, str):
            tf_padding = self.padding.upper()
        else:
            tf_padding = self.padding

        return tf.nn.convolution(
            inputs,
            kernel,
            strides=list(self.strides),
            padding=tf_padding,
            dilations=list(self.dilation_rate),
            data_format=self._tf_data_format,
            name=self.__class__.__name__,
        )

    # TODO(b/213173659): remove this when grouped convolutions are fully
    # supported on the CPU for compiled functions. For now, we need this as a
    # workaround for CPU support.
    @tf.function(jit_compile=True)
    def _jit_compiled_convolution_op(self, inputs, kernel):
        return self.convolution_op(inputs, kernel)

    def call(self, inputs):
        input_shape = inputs.shape

        if self._is_causal:  # Apply causal padding to inputs for Conv1D.
            inputs = tf.pad(inputs, self._compute_causal_padding(inputs))

        if self.groups > 1:
            outputs = self._jit_compiled_convolution_op(
                inputs, tf.convert_to_tensor(self.kernel)
            )
        else:
            outputs = self.convolution_op(inputs, self.kernel)

        if self.use_bias:
            output_rank = outputs.shape.rank
            if self.rank == 1 and self._channels_first:
                # nn.bias_add does not accept a 1D input tensor.
                bias = tf.reshape(self.bias, (1, self.filters, 1))
                outputs += bias
            else:
                # Handle multiple batch dimensions.
                if output_rank is not None and output_rank > 2 + self.rank:

                    def _apply_fn(o):
                        return tf.nn.bias_add(
                            o, self.bias, data_format=self._tf_data_format
                        )

                    outputs = conv_utils.squeeze_batch_dims(
                        outputs, _apply_fn, inner_rank=self.rank + 1
                    )
                else:
                    outputs = tf.nn.bias_add(
                        outputs, self.bias, data_format=self._tf_data_format
                    )

        if not tf.executing_eagerly() and input_shape.rank:
            # Infer the static output shape:
            out_shape = self.compute_output_shape(input_shape)
            outputs.set_shape(out_shape)

        if self.activation is not None:
            return self.activation(outputs)
        return outputs

    def _spatial_output_shape(self, spatial_input_shape):
        return [
            conv_utils.conv_output_length(
                length,
                self.kernel_size[i],
                padding=self.padding,
                stride=self.strides[i],
                dilation=self.dilation_rate[i],
            )
            for i, length in enumerate(spatial_input_shape)
        ]

    def compute_output_shape(self, input_shape):
        input_shape = tf.TensorShape(input_shape).as_list()
        batch_rank = len(input_shape) - self.rank - 1
        try:
            if self.data_format == "channels_last":
                return tf.TensorShape(
                    input_shape[:batch_rank]
                    + self._spatial_output_shape(input_shape[batch_rank:-1])
                    + [self.filters]
                )
            else:
                return tf.TensorShape(
                    input_shape[:batch_rank]
                    + [self.filters]
                    + self._spatial_output_shape(input_shape[batch_rank + 1 :])
                )

        except ValueError:
            raise ValueError(
                "One of the dimensions in the output is <= 0 "
                f"due to downsampling in {self.name}. Consider "
                "increasing the input size. "
                f"Received input shape {input_shape} which would produce "
                "output shape with a zero or negative value in a "
                "dimension."
            )

    def _recreate_conv_op(self, inputs):
        return False

    def get_config(self):
        config = {
            "filters": self.filters,
            "kernel_size": self.kernel_size,
            "strides": self.strides,
            "padding": self.padding,
            "data_format": self.data_format,
            "dilation_rate": self.dilation_rate,
            "groups": self.groups,
            "activation": activations.serialize(self.activation),
            "use_bias": self.use_bias,
            "kernel_initializer": initializers.serialize(
                self.kernel_initializer
            ),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "kernel_regularizer": regularizers.serialize(
                self.kernel_regularizer
            ),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
            "activity_regularizer": regularizers.serialize(
                self.activity_regularizer
            ),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "bias_constraint": constraints.serialize(self.bias_constraint),
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def _compute_causal_padding(self, inputs):
        """Calculates padding for 'causal' option for 1-d conv layers."""
        left_pad = self.dilation_rate[0] * (self.kernel_size[0] - 1)
        if getattr(inputs.shape, "ndims", None) is None:
            batch_rank = 1
        else:
            batch_rank = len(inputs.shape) - 2
        if self.data_format == "channels_last":
            causal_padding = [[0, 0]] * batch_rank + [[left_pad, 0], [0, 0]]
        else:
            causal_padding = [[0, 0]] * batch_rank + [[0, 0], [left_pad, 0]]
        return causal_padding

    def _get_channel_axis(self):
        if self.data_format == "channels_first":
            return -1 - self.rank
        else:
            return -1

    def _get_input_channel(self, input_shape):
        channel_axis = self._get_channel_axis()
        if input_shape.dims[channel_axis].value is None:
            raise ValueError(
                "The channel dimension of the inputs should be defined. "
                f"The input_shape received is {input_shape}, "
                f"where axis {channel_axis} (0-based) "
                "is the channel dimension, which found to be `None`."
            )
        return int(input_shape[channel_axis])

    def _get_padding_op(self):
        if self.padding == "causal":
            op_padding = "valid"
        else:
            op_padding = self.padding
        if not isinstance(op_padding, (list, tuple)):
            op_padding = op_padding.upper()
        return op_padding



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.keras.saving namespace.
"""

import sys as _sys

from keras.saving.object_registration import CustomObjectScope as custom_object_scope
from keras.saving.object_registration import get_custom_objects
from keras.saving.object_registration import get_registered_name
from keras.saving.object_registration import get_registered_object
from keras.saving.object_registration import register_keras_serializable
from keras.saving.saving_api import load_model
from keras.saving.saving_api import save_model


# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities used by convolution layers."""

import itertools

import numpy as np
import tensorflow.compat.v2 as tf

from keras import backend


def convert_data_format(data_format, ndim):
    if data_format == "channels_last":
        if ndim == 3:
            return "NWC"
        elif ndim == 4:
            return "NHWC"
        elif ndim == 5:
            return "NDHWC"
        else:
            raise ValueError(
                f"Input rank not supported: {ndim}. "
                "Expected values are [3, 4, 5]"
            )
    elif data_format == "channels_first":
        if ndim == 3:
            return "NCW"
        elif ndim == 4:
            return "NCHW"
        elif ndim == 5:
            return "NCDHW"
        else:
            raise ValueError(
                f"Input rank not supported: {ndim}. "
                "Expected values are [3, 4, 5]"
            )
    else:
        raise ValueError(
            f"Invalid data_format: {data_format}. "
            'Expected values are ["channels_first", "channels_last"]'
        )


def normalize_tuple(value, n, name, allow_zero=False):
    """Transforms non-negative/positive integer/integers into an integer tuple.

    Args:
      value: The value to validate and convert. Could an int, or any iterable of
        ints.
      n: The size of the tuple to be returned.
      name: The name of the argument being validated, e.g. "strides" or
        "kernel_size". This is only used to format error messages.
      allow_zero: Default to False. A ValueError will raised if zero is received
        and this param is False.

    Returns:
      A tuple of n integers.

    Raises:
      ValueError: If something else than an int/long or iterable thereof or a
      negative value is
        passed.
    """
    error_msg = (
        f"The `{name}` argument must be a tuple of {n} "
        f"integers. Received: {value}"
    )

    if isinstance(value, int):
        value_tuple = (value,) * n
    else:
        try:
            value_tuple = tuple(value)
        except TypeError:
            raise ValueError(error_msg)
        if len(value_tuple) != n:
            raise ValueError(error_msg)
        for single_value in value_tuple:
            try:
                int(single_value)
            except (ValueError, TypeError):
                error_msg += (
                    f"including element {single_value} of "
                    f"type {type(single_value)}"
                )
                raise ValueError(error_msg)

    if allow_zero:
        unqualified_values = {v for v in value_tuple if v < 0}
        req_msg = ">= 0"
    else:
        unqualified_values = {v for v in value_tuple if v <= 0}
        req_msg = "> 0"

    if unqualified_values:
        error_msg += (
            f" including {unqualified_values}"
            f" that does not satisfy the requirement `{req_msg}`."
        )
        raise ValueError(error_msg)

    return value_tuple


def conv_output_length(input_length, filter_size, padding, stride, dilation=1):
    """Determines output length of a convolution given input length.

    Args:
        input_length: integer.
        filter_size: integer.
        padding: one of "same", "valid", "full", "causal"
        stride: integer.
        dilation: dilation rate, integer.

    Returns:
        The output length (integer).
    """
    if input_length is None:
        return None
    assert padding in {"same", "valid", "full", "causal"}
    dilated_filter_size = filter_size + (filter_size - 1) * (dilation - 1)
    if padding in ["same", "causal"]:
        output_length = input_length
    elif padding == "valid":
        output_length = input_length - dilated_filter_size + 1
    elif padding == "full":
        output_length = input_length + dilated_filter_size - 1
    return (output_length + stride - 1) // stride


def conv_input_length(output_length, filter_size, padding, stride):
    """Determines input length of a convolution given output length.

    Args:
        output_length: integer.
        filter_size: integer.
        padding: one of "same", "valid", "full".
        stride: integer.

    Returns:
        The input length (integer).
    """
    if output_length is None:
        return None
    assert padding in {"same", "valid", "full"}
    if padding == "same":
        pad = filter_size // 2
    elif padding == "valid":
        pad = 0
    elif padding == "full":
        pad = filter_size - 1
    return (output_length - 1) * stride - 2 * pad + filter_size


def deconv_output_length(
    input_length,
    filter_size,
    padding,
    output_padding=None,
    stride=0,
    dilation=1,
):
    """Determines output length of a transposed convolution given input length.

    Args:
        input_length: Integer.
        filter_size: Integer.
        padding: one of `"same"`, `"valid"`, `"full"`.
        output_padding: Integer, amount of padding along the output dimension.
          Can be set to `None` in which case the output length is inferred.
        stride: Integer.
        dilation: Integer.

    Returns:
        The output length (integer).
    """
    assert padding in {"same", "valid", "full"}
    if input_length is None:
        return None

    # Get the dilated kernel size
    filter_size = filter_size + (filter_size - 1) * (dilation - 1)

    # Infer length if output padding is None, else compute the exact length
    if output_padding is None:
        if padding == "valid":
            length = input_length * stride + max(filter_size - stride, 0)
        elif padding == "full":
            length = input_length * stride - (stride + filter_size - 2)
        elif padding == "same":
            length = input_length * stride

    else:
        if padding == "same":
            pad = filter_size // 2
        elif padding == "valid":
            pad = 0
        elif padding == "full":
            pad = filter_size - 1

        length = (
            (input_length - 1) * stride + filter_size - 2 * pad + output_padding
        )
    return length


def normalize_data_format(value):
    if value is None:
        value = backend.image_data_format()
    data_format = value.lower()
    if data_format not in {"channels_first", "channels_last"}:
        raise ValueError(
            "The `data_format` argument must be one of "
            f'"channels_first", "channels_last". Received: {value}'
        )
    return data_format


def normalize_padding(value):
    if isinstance(value, (list, tuple)):
        return value
    padding = value.lower()
    if padding not in {"valid", "same", "causal"}:
        raise ValueError(
            "The `padding` argument must be a list/tuple or one of "
            '"valid", "same" (or "causal", only for `Conv1D). '
            f"Received: {padding}"
        )
    return padding


def conv_kernel_mask(input_shape, kernel_shape, strides, padding):
    """Compute a mask representing the connectivity of a convolution operation.

    Assume a convolution with given parameters is applied to an input having N
    spatial dimensions with `input_shape = (d_in1, ..., d_inN)` to produce an
    output with shape `(d_out1, ..., d_outN)`. This method returns a boolean
    array of shape `(d_in1, ..., d_inN, d_out1, ..., d_outN)` with `True`
    entries indicating pairs of input and output locations that are connected by
    a weight.

    Example:

      >>> input_shape = (4,)
      >>> kernel_shape = (2,)
      >>> strides = (1,)
      >>> padding = "valid"
      >>> conv_kernel_mask(input_shape, kernel_shape, strides, padding)
      array([[ True, False, False],
             [ True,  True, False],
             [False,  True,  True],
             [False, False,  True]])

      where rows and columns correspond to inputs and outputs respectively.


    Args:
      input_shape: tuple of size N: `(d_in1, ..., d_inN)`, spatial shape of the
        input.
      kernel_shape: tuple of size N, spatial shape of the convolutional kernel /
        receptive field.
      strides: tuple of size N, strides along each spatial dimension.
      padding: type of padding, string `"same"` or `"valid"`.
        `"valid"` means no padding. `"same"` results in padding evenly to
        the left/right or up/down of the input such that output has the same
        height/width dimension as the input.

    Returns:
      A boolean 2N-D `np.ndarray` of shape
      `(d_in1, ..., d_inN, d_out1, ..., d_outN)`, where `(d_out1, ..., d_outN)`
      is the spatial shape of the output. `True` entries in the mask represent
      pairs of input-output locations that are connected by a weight.

    Raises:
      ValueError: if `input_shape`, `kernel_shape` and `strides` don't have the
          same number of dimensions.
      NotImplementedError: if `padding` is not in {`"same"`, `"valid"`}.
    """
    if padding not in {"same", "valid"}:
        raise NotImplementedError(
            f"Padding type {padding} not supported. "
            'Only "valid" and "same" are implemented.'
        )

    in_dims = len(input_shape)
    if isinstance(kernel_shape, int):
        kernel_shape = (kernel_shape,) * in_dims
    if isinstance(strides, int):
        strides = (strides,) * in_dims

    kernel_dims = len(kernel_shape)
    stride_dims = len(strides)
    if kernel_dims != in_dims or stride_dims != in_dims:
        raise ValueError(
            "Number of strides, input and kernel dimensions must all "
            f"match. Received: stride_dims={stride_dims}, "
            f"in_dims={in_dims}, kernel_dims={kernel_dims}"
        )

    output_shape = conv_output_shape(
        input_shape, kernel_shape, strides, padding
    )

    mask_shape = input_shape + output_shape
    mask = np.zeros(mask_shape, bool)

    output_axes_ticks = [range(dim) for dim in output_shape]
    for output_position in itertools.product(*output_axes_ticks):
        input_axes_ticks = conv_connected_inputs(
            input_shape, kernel_shape, output_position, strides, padding
        )
        for input_position in itertools.product(*input_axes_ticks):
            mask[input_position + output_position] = True

    return mask


def conv_kernel_idxs(
    input_shape,
    kernel_shape,
    strides,
    padding,
    filters_in,
    filters_out,
    data_format,
):
    """Yields output-input tuples of indices in a CNN layer.

    The generator iterates over all `(output_idx, input_idx)` tuples, where
    `output_idx` is an integer index in a flattened tensor representing a single
    output image of a convolutional layer that is connected (via the layer
    weights) to the respective single input image at `input_idx`

    Example:

      >>> input_shape = (2, 2)
      >>> kernel_shape = (2, 1)
      >>> strides = (1, 1)
      >>> padding = "valid"
      >>> filters_in = 1
      >>> filters_out = 1
      >>> data_format = "channels_last"
      >>> list(conv_kernel_idxs(input_shape, kernel_shape, strides, padding,
      ...                       filters_in, filters_out, data_format))
      [(0, 0), (0, 2), (1, 1), (1, 3)]

    Args:
      input_shape: tuple of size N: `(d_in1, ..., d_inN)`, spatial shape of the
        input.
      kernel_shape: tuple of size N, spatial shape of the convolutional kernel /
        receptive field.
      strides: tuple of size N, strides along each spatial dimension.
      padding: type of padding, string `"same"` or `"valid"`.
        `"valid"` means no padding. `"same"` results in padding evenly to
        the left/right or up/down of the input such that output has the same
        height/width dimension as the input.
      filters_in: `int`, number if filters in the input to the layer.
      filters_out: `int', number if filters in the output of the layer.
      data_format: string, "channels_first" or "channels_last".

    Yields:
      The next tuple `(output_idx, input_idx)`, where `output_idx` is an integer
      index in a flattened tensor representing a single output image of a
      convolutional layer that is connected (via the layer weights) to the
      respective single input image at `input_idx`.

    Raises:
        ValueError: if `data_format` is neither `"channels_last"` nor
          `"channels_first"`, or if number of strides, input, and kernel number
          of dimensions do not match.

        NotImplementedError: if `padding` is neither `"same"` nor `"valid"`.
    """
    if padding not in ("same", "valid"):
        raise NotImplementedError(
            f"Padding type {padding} not supported. "
            'Only "valid" and "same" are implemented.'
        )

    in_dims = len(input_shape)
    if isinstance(kernel_shape, int):
        kernel_shape = (kernel_shape,) * in_dims
    if isinstance(strides, int):
        strides = (strides,) * in_dims

    kernel_dims = len(kernel_shape)
    stride_dims = len(strides)
    if kernel_dims != in_dims or stride_dims != in_dims:
        raise ValueError(
            "Number of strides, input and kernel dimensions must all "
            f"match. Received: stride_dims={stride_dims}, "
            f"in_dims={in_dims}, kernel_dims={kernel_dims}"
        )

    output_shape = conv_output_shape(
        input_shape, kernel_shape, strides, padding
    )
    output_axes_ticks = [range(dim) for dim in output_shape]

    if data_format == "channels_first":
        concat_idxs = (
            lambda spatial_idx, filter_idx: (filter_idx,) + spatial_idx
        )
    elif data_format == "channels_last":
        concat_idxs = lambda spatial_idx, filter_idx: spatial_idx + (
            filter_idx,
        )
    else:
        raise ValueError(
            f"Data format `{data_format}` not recognized."
            '`data_format` must be "channels_first" or "channels_last".'
        )

    for output_position in itertools.product(*output_axes_ticks):
        input_axes_ticks = conv_connected_inputs(
            input_shape, kernel_shape, output_position, strides, padding
        )
        for input_position in itertools.product(*input_axes_ticks):
            for f_in in range(filters_in):
                for f_out in range(filters_out):
                    out_idx = np.ravel_multi_index(
                        multi_index=concat_idxs(output_position, f_out),
                        dims=concat_idxs(output_shape, filters_out),
                    )
                    in_idx = np.ravel_multi_index(
                        multi_index=concat_idxs(input_position, f_in),
                        dims=concat_idxs(input_shape, filters_in),
                    )
                    yield (out_idx, in_idx)


def conv_connected_inputs(
    input_shape, kernel_shape, output_position, strides, padding
):
    """Return locations of the input connected to an output position.

    Assume a convolution with given parameters is applied to an input having N
    spatial dimensions with `input_shape = (d_in1, ..., d_inN)`. This method
    returns N ranges specifying the input region that was convolved with the
    kernel to produce the output at position
    `output_position = (p_out1, ..., p_outN)`.

    Example:

      >>> input_shape = (4, 4)
      >>> kernel_shape = (2, 1)
      >>> output_position = (1, 1)
      >>> strides = (1, 1)
      >>> padding = "valid"
      >>> conv_connected_inputs(input_shape, kernel_shape, output_position,
      ...                       strides, padding)
      [range(1, 3), range(1, 2)]

    Args:
      input_shape: tuple of size N: `(d_in1, ..., d_inN)`, spatial shape of the
        input.
      kernel_shape: tuple of size N, spatial shape of the convolutional kernel /
        receptive field.
      output_position: tuple of size N: `(p_out1, ..., p_outN)`, a single
        position in the output of the convolution.
      strides: tuple of size N, strides along each spatial dimension.
      padding: type of padding, string `"same"` or `"valid"`.
        `"valid"` means no padding. `"same"` results in padding evenly to
        the left/right or up/down of the input such that output has the same
        height/width dimension as the input.

    Returns:
      N ranges `[[p_in_left1, ..., p_in_right1], ...,
                [p_in_leftN, ..., p_in_rightN]]` specifying the region in the
      input connected to output_position.
    """
    ranges = []

    ndims = len(input_shape)
    for d in range(ndims):
        left_shift = int(kernel_shape[d] / 2)
        right_shift = kernel_shape[d] - left_shift

        center = output_position[d] * strides[d]

        if padding == "valid":
            center += left_shift

        start = max(0, center - left_shift)
        end = min(input_shape[d], center + right_shift)

        ranges.append(range(start, end))

    return ranges


def conv_output_shape(input_shape, kernel_shape, strides, padding):
    """Return the output shape of an N-D convolution.

    Forces dimensions where input is empty (size 0) to remain empty.

    Args:
      input_shape: tuple of size N: `(d_in1, ..., d_inN)`, spatial shape of the
        input.
      kernel_shape: tuple of size N, spatial shape of the convolutional kernel /
        receptive field.
      strides: tuple of size N, strides along each spatial dimension.
      padding: type of padding, string `"same"` or `"valid"`.
        `"valid"` means no padding. `"same"` results in padding evenly to
        the left/right or up/down of the input such that output has the same
        height/width dimension as the input.

    Returns:
      tuple of size N: `(d_out1, ..., d_outN)`, spatial shape of the output.
    """
    dims = range(len(kernel_shape))
    output_shape = [
        conv_output_length(input_shape[d], kernel_shape[d], padding, strides[d])
        for d in dims
    ]
    output_shape = tuple(
        [0 if input_shape[d] == 0 else output_shape[d] for d in dims]
    )
    return output_shape


def squeeze_batch_dims(inp, op, inner_rank):
    """Returns `unsqueeze_batch(op(squeeze_batch(inp)))`.

    Where `squeeze_batch` reshapes `inp` to shape
    `[prod(inp.shape[:-inner_rank])] + inp.shape[-inner_rank:]`
    and `unsqueeze_batch` does the reverse reshape but on the output.

    Args:
      inp: A tensor with dims `batch_shape + inner_shape` where `inner_shape`
        is length `inner_rank`.
      op: A callable that takes a single input tensor and returns a single.
        output tensor.
      inner_rank: A python integer.

    Returns:
      `unsqueeze_batch_op(squeeze_batch(inp))`.
    """
    with tf.name_scope("squeeze_batch_dims"):
        shape = inp.shape

        inner_shape = shape[-inner_rank:]
        if not inner_shape.is_fully_defined():
            inner_shape = tf.shape(inp)[-inner_rank:]

        batch_shape = shape[:-inner_rank]
        if not batch_shape.is_fully_defined():
            batch_shape = tf.shape(inp)[:-inner_rank]

        if isinstance(inner_shape, tf.TensorShape):
            inp_reshaped = tf.reshape(inp, [-1] + inner_shape.as_list())
        else:
            inp_reshaped = tf.reshape(
                inp, tf.concat(([-1], inner_shape), axis=-1)
            )

        out_reshaped = op(inp_reshaped)

        out_inner_shape = out_reshaped.shape[-inner_rank:]
        if not out_inner_shape.is_fully_defined():
            out_inner_shape = tf.shape(out_reshaped)[-inner_rank:]

        out = tf.reshape(
            out_reshaped, tf.concat((batch_shape, out_inner_shape), axis=-1)
        )

        out.set_shape(inp.shape[:-inner_rank] + out.shape[-inner_rank:])
        return out



import tensorflow as tf
from tensorflow import keras

from keras.integration_test.models.input_spec import InputSpec

TIMESTEPS = 64
INPUT_DIM = 50
OUTPUT_DIM = 40
NUM_RNN_LAYERS = 2
RNN_UNITS = 32


def get_input_preprocessor():
    return None


def get_data_spec(batch_size):
    return (
        InputSpec((batch_size, TIMESTEPS, INPUT_DIM)),
        InputSpec((batch_size, 1), dtype="int64", range=[0, OUTPUT_DIM]),
    )


def ctc_loss(y_true, y_pred):
    batch_length = tf.cast(tf.shape(y_true)[0], dtype="int64")
    input_length = tf.cast(tf.shape(y_pred)[1], dtype="int64")
    label_length = tf.cast(tf.shape(y_true)[1], dtype="int64")

    input_length = input_length * tf.ones(
        shape=(batch_length, 1), dtype="int64"
    )
    label_length = label_length * tf.ones(
        shape=(batch_length, 1), dtype="int64"
    )

    return keras.backend.ctc_batch_cost(
        y_true, y_pred, input_length, label_length
    )


def get_model(
    build=False, compile=False, jit_compile=False, include_preprocessing=True
):
    input_spectrogram = keras.layers.Input((None, INPUT_DIM), name="input")
    x = keras.layers.Reshape((-1, INPUT_DIM, 1), name="expand_dim")(
        input_spectrogram
    )
    x = keras.layers.Conv2D(
        filters=32,
        kernel_size=[11, 41],
        strides=[2, 2],
        padding="same",
        use_bias=False,
        name="conv_1",
    )(x)
    x = keras.layers.BatchNormalization(name="conv_1_bn")(x)
    x = keras.layers.ReLU(name="conv_1_relu")(x)
    x = keras.layers.Conv2D(
        filters=32,
        kernel_size=[11, 21],
        strides=[1, 2],
        padding="same",
        use_bias=False,
        name="conv_2",
    )(x)
    x = keras.layers.BatchNormalization(name="conv_2_bn")(x)
    x = keras.layers.ReLU(name="conv_2_relu")(x)
    x = keras.layers.Reshape((-1, x.shape[-2] * x.shape[-1]))(x)
    for i in range(1, NUM_RNN_LAYERS + 1):
        recurrent = keras.layers.GRU(
            units=RNN_UNITS,
            activation="tanh",
            recurrent_activation="sigmoid",
            use_bias=True,
            return_sequences=True,
            reset_after=True,
            name=f"gru_{i}",
        )
        x = keras.layers.Bidirectional(
            recurrent, name=f"bidirectional_{i}", merge_mode="concat"
        )(x)
        if i < NUM_RNN_LAYERS:
            x = keras.layers.Dropout(rate=0.5)(x)
    x = keras.layers.Dense(units=RNN_UNITS * 2, name="dense_1")(x)
    x = keras.layers.ReLU(name="dense_1_relu")(x)
    x = keras.layers.Dropout(rate=0.5)(x)
    output = keras.layers.Dense(units=OUTPUT_DIM + 1, activation="softmax")(x)
    model = keras.Model(input_spectrogram, output, name="DeepSpeech_2")

    if compile:
        model.compile(
            optimizer=keras.optimizers.Adam(learning_rate=1e-4),
            loss=ctc_loss,
            jit_compile=jit_compile,
        )
    return model


def get_custom_objects():
    return {"ctc_loss": ctc_loss}



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Long Short-Term Memory layer."""


import uuid

import tensorflow.compat.v2 as tf

from keras import activations
from keras import backend
from keras import constraints
from keras import initializers
from keras import regularizers
from keras.engine import base_layer
from keras.engine.input_spec import InputSpec
from keras.layers.rnn import gru_lstm_utils
from keras.layers.rnn import rnn_utils
from keras.layers.rnn.base_rnn import RNN
from keras.layers.rnn.dropout_rnn_cell_mixin import DropoutRNNCellMixin
from keras.utils import tf_utils

# isort: off
from tensorflow.python.platform import tf_logging as logging
from tensorflow.python.util.tf_export import keras_export

RECURRENT_DROPOUT_WARNING_MSG = (
    "RNN `implementation=2` is not supported when `recurrent_dropout` is set. "
    "Using `implementation=1`."
)


@keras_export("keras.layers.LSTMCell", v1=[])
class LSTMCell(DropoutRNNCellMixin, base_layer.BaseRandomLayer):
    """Cell class for the LSTM layer.

    See [the Keras RNN API guide](https://www.tensorflow.org/guide/keras/rnn)
    for details about the usage of RNN API.

    This class processes one step within the whole time sequence input, whereas
    `tf.keras.layer.LSTM` processes the whole sequence.

    For example:

    >>> inputs = tf.random.normal([32, 10, 8])
    >>> rnn = tf.keras.layers.RNN(tf.keras.layers.LSTMCell(4))
    >>> output = rnn(inputs)
    >>> print(output.shape)
    (32, 4)
    >>> rnn = tf.keras.layers.RNN(
    ...    tf.keras.layers.LSTMCell(4),
    ...    return_sequences=True,
    ...    return_state=True)
    >>> whole_seq_output, final_memory_state, final_carry_state = rnn(inputs)
    >>> print(whole_seq_output.shape)
    (32, 10, 4)
    >>> print(final_memory_state.shape)
    (32, 4)
    >>> print(final_carry_state.shape)
    (32, 4)

    Args:
      units: Positive integer, dimensionality of the output space.
      activation: Activation function to use. Default: hyperbolic tangent
        (`tanh`). If you pass `None`, no activation is applied (ie. "linear"
        activation: `a(x) = x`).
      recurrent_activation: Activation function to use for the recurrent step.
        Default: sigmoid (`sigmoid`). If you pass `None`, no activation is
        applied (ie. "linear" activation: `a(x) = x`).
      use_bias: Boolean, (default `True`), whether the layer uses a bias vector.
      kernel_initializer: Initializer for the `kernel` weights matrix, used for
        the linear transformation of the inputs. Default: `glorot_uniform`.
      recurrent_initializer: Initializer for the `recurrent_kernel` weights
        matrix, used for the linear transformation of the recurrent state.
        Default: `orthogonal`.
      bias_initializer: Initializer for the bias vector. Default: `zeros`.
      unit_forget_bias: Boolean (default `True`). If True, add 1 to the bias of
        the forget gate at initialization. Setting it to true will also force
        `bias_initializer="zeros"`. This is recommended in [Jozefowicz et
          al.](https://github.com/mlresearch/v37/blob/gh-pages/jozefowicz15.pdf)
      kernel_regularizer: Regularizer function applied to the `kernel` weights
        matrix. Default: `None`.
      recurrent_regularizer: Regularizer function applied to
        the `recurrent_kernel` weights matrix. Default: `None`.
      bias_regularizer: Regularizer function applied to the bias vector.
        Default: `None`.
      kernel_constraint: Constraint function applied to the `kernel` weights
        matrix. Default: `None`.
      recurrent_constraint: Constraint function applied to the
        `recurrent_kernel` weights matrix. Default: `None`.
      bias_constraint: Constraint function applied to the bias vector. Default:
        `None`.
      dropout: Float between 0 and 1. Fraction of the units to drop for the
        linear transformation of the inputs. Default: 0.
      recurrent_dropout: Float between 0 and 1. Fraction of the units to drop
        for the linear transformation of the recurrent state. Default: 0.

    Call arguments:
      inputs: A 2D tensor, with shape of `[batch, feature]`.
      states: List of 2 tensors that corresponding to the cell's units. Both of
        them have shape `[batch, units]`, the first tensor is the memory state
        from previous time step, the second tensor is the carry state from
        previous time step. For timestep 0, the initial state provided by user
        will be feed to cell.
      training: Python boolean indicating whether the layer should behave in
        training mode or in inference mode. Only relevant when `dropout` or
        `recurrent_dropout` is used.
    """

    def __init__(
        self,
        units,
        activation="tanh",
        recurrent_activation="sigmoid",
        use_bias=True,
        kernel_initializer="glorot_uniform",
        recurrent_initializer="orthogonal",
        bias_initializer="zeros",
        unit_forget_bias=True,
        kernel_regularizer=None,
        recurrent_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        recurrent_constraint=None,
        bias_constraint=None,
        dropout=0.0,
        recurrent_dropout=0.0,
        **kwargs,
    ):
        if units <= 0:
            raise ValueError(
                "Received an invalid value for argument `units`, "
                f"expected a positive integer, got {units}."
            )
        # By default use cached variable under v2 mode, see b/143699808.
        if tf.compat.v1.executing_eagerly_outside_functions():
            self._enable_caching_device = kwargs.pop(
                "enable_caching_device", True
            )
        else:
            self._enable_caching_device = kwargs.pop(
                "enable_caching_device", False
            )
        super().__init__(**kwargs)
        self.units = units
        self.activation = activations.get(activation)
        self.recurrent_activation = activations.get(recurrent_activation)
        self.use_bias = use_bias

        self.kernel_initializer = initializers.get(kernel_initializer)
        self.recurrent_initializer = initializers.get(recurrent_initializer)
        self.bias_initializer = initializers.get(bias_initializer)
        self.unit_forget_bias = unit_forget_bias

        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.recurrent_regularizer = regularizers.get(recurrent_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)

        self.kernel_constraint = constraints.get(kernel_constraint)
        self.recurrent_constraint = constraints.get(recurrent_constraint)
        self.bias_constraint = constraints.get(bias_constraint)

        self.dropout = min(1.0, max(0.0, dropout))
        self.recurrent_dropout = min(1.0, max(0.0, recurrent_dropout))
        implementation = kwargs.pop("implementation", 2)
        if self.recurrent_dropout != 0 and implementation != 1:
            logging.debug(RECURRENT_DROPOUT_WARNING_MSG)
            self.implementation = 1
        else:
            self.implementation = implementation
        self.state_size = [self.units, self.units]
        self.output_size = self.units

    @tf_utils.shape_type_conversion
    def build(self, input_shape):
        super().build(input_shape)
        default_caching_device = rnn_utils.caching_device(self)
        input_dim = input_shape[-1]
        self.kernel = self.add_weight(
            shape=(input_dim, self.units * 4),
            name="kernel",
            initializer=self.kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
            caching_device=default_caching_device,
        )
        self.recurrent_kernel = self.add_weight(
            shape=(self.units, self.units * 4),
            name="recurrent_kernel",
            initializer=self.recurrent_initializer,
            regularizer=self.recurrent_regularizer,
            constraint=self.recurrent_constraint,
            caching_device=default_caching_device,
        )

        if self.use_bias:
            if self.unit_forget_bias:

                def bias_initializer(_, *args, **kwargs):
                    return backend.concatenate(
                        [
                            self.bias_initializer(
                                (self.units,), *args, **kwargs
                            ),
                            initializers.get("ones")(
                                (self.units,), *args, **kwargs
                            ),
                            self.bias_initializer(
                                (self.units * 2,), *args, **kwargs
                            ),
                        ]
                    )

            else:
                bias_initializer = self.bias_initializer
            self.bias = self.add_weight(
                shape=(self.units * 4,),
                name="bias",
                initializer=bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                caching_device=default_caching_device,
            )
        else:
            self.bias = None
        self.built = True

    def _compute_carry_and_output(self, x, h_tm1, c_tm1):
        """Computes carry and output using split kernels."""
        x_i, x_f, x_c, x_o = x
        h_tm1_i, h_tm1_f, h_tm1_c, h_tm1_o = h_tm1
        i = self.recurrent_activation(
            x_i + backend.dot(h_tm1_i, self.recurrent_kernel[:, : self.units])
        )
        f = self.recurrent_activation(
            x_f
            + backend.dot(
                h_tm1_f, self.recurrent_kernel[:, self.units : self.units * 2]
            )
        )
        c = f * c_tm1 + i * self.activation(
            x_c
            + backend.dot(
                h_tm1_c,
                self.recurrent_kernel[:, self.units * 2 : self.units * 3],
            )
        )
        o = self.recurrent_activation(
            x_o
            + backend.dot(h_tm1_o, self.recurrent_kernel[:, self.units * 3 :])
        )
        return c, o

    def _compute_carry_and_output_fused(self, z, c_tm1):
        """Computes carry and output using fused kernels."""
        z0, z1, z2, z3 = z
        i = self.recurrent_activation(z0)
        f = self.recurrent_activation(z1)
        c = f * c_tm1 + i * self.activation(z2)
        o = self.recurrent_activation(z3)
        return c, o

    def call(self, inputs, states, training=None):
        h_tm1 = states[0]  # previous memory state
        c_tm1 = states[1]  # previous carry state

        dp_mask = self.get_dropout_mask_for_cell(inputs, training, count=4)
        rec_dp_mask = self.get_recurrent_dropout_mask_for_cell(
            h_tm1, training, count=4
        )

        if self.implementation == 1:
            if 0 < self.dropout < 1.0:
                inputs_i = inputs * dp_mask[0]
                inputs_f = inputs * dp_mask[1]
                inputs_c = inputs * dp_mask[2]
                inputs_o = inputs * dp_mask[3]
            else:
                inputs_i = inputs
                inputs_f = inputs
                inputs_c = inputs
                inputs_o = inputs
            k_i, k_f, k_c, k_o = tf.split(
                self.kernel, num_or_size_splits=4, axis=1
            )
            x_i = backend.dot(inputs_i, k_i)
            x_f = backend.dot(inputs_f, k_f)
            x_c = backend.dot(inputs_c, k_c)
            x_o = backend.dot(inputs_o, k_o)
            if self.use_bias:
                b_i, b_f, b_c, b_o = tf.split(
                    self.bias, num_or_size_splits=4, axis=0
                )
                x_i = backend.bias_add(x_i, b_i)
                x_f = backend.bias_add(x_f, b_f)
                x_c = backend.bias_add(x_c, b_c)
                x_o = backend.bias_add(x_o, b_o)

            if 0 < self.recurrent_dropout < 1.0:
                h_tm1_i = h_tm1 * rec_dp_mask[0]
                h_tm1_f = h_tm1 * rec_dp_mask[1]
                h_tm1_c = h_tm1 * rec_dp_mask[2]
                h_tm1_o = h_tm1 * rec_dp_mask[3]
            else:
                h_tm1_i = h_tm1
                h_tm1_f = h_tm1
                h_tm1_c = h_tm1
                h_tm1_o = h_tm1
            x = (x_i, x_f, x_c, x_o)
            h_tm1 = (h_tm1_i, h_tm1_f, h_tm1_c, h_tm1_o)
            c, o = self._compute_carry_and_output(x, h_tm1, c_tm1)
        else:
            if 0.0 < self.dropout < 1.0:
                inputs = inputs * dp_mask[0]
            z = backend.dot(inputs, self.kernel)
            z += backend.dot(h_tm1, self.recurrent_kernel)
            if self.use_bias:
                z = backend.bias_add(z, self.bias)

            z = tf.split(z, num_or_size_splits=4, axis=1)
            c, o = self._compute_carry_and_output_fused(z, c_tm1)

        h = o * self.activation(c)
        return h, [h, c]

    def get_config(self):
        config = {
            "units": self.units,
            "activation": activations.serialize(self.activation),
            "recurrent_activation": activations.serialize(
                self.recurrent_activation
            ),
            "use_bias": self.use_bias,
            "kernel_initializer": initializers.serialize(
                self.kernel_initializer
            ),
            "recurrent_initializer": initializers.serialize(
                self.recurrent_initializer
            ),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "unit_forget_bias": self.unit_forget_bias,
            "kernel_regularizer": regularizers.serialize(
                self.kernel_regularizer
            ),
            "recurrent_regularizer": regularizers.serialize(
                self.recurrent_regularizer
            ),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "recurrent_constraint": constraints.serialize(
                self.recurrent_constraint
            ),
            "bias_constraint": constraints.serialize(self.bias_constraint),
            "dropout": self.dropout,
            "recurrent_dropout": self.recurrent_dropout,
            "implementation": self.implementation,
        }
        config.update(rnn_utils.config_for_enable_caching_device(self))
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def get_initial_state(self, inputs=None, batch_size=None, dtype=None):
        return list(
            rnn_utils.generate_zero_filled_state_for_cell(
                self, inputs, batch_size, dtype
            )
        )


@keras_export("keras.layers.LSTM", v1=[])
class LSTM(DropoutRNNCellMixin, RNN, base_layer.BaseRandomLayer):
    """Long Short-Term Memory layer - Hochreiter 1997.

    See [the Keras RNN API guide](https://www.tensorflow.org/guide/keras/rnn)
    for details about the usage of RNN API.

    Based on available runtime hardware and constraints, this layer
    will choose different implementations (cuDNN-based or pure-TensorFlow)
    to maximize the performance. If a GPU is available and all
    the arguments to the layer meet the requirement of the cuDNN kernel
    (see below for details), the layer will use a fast cuDNN implementation.

    The requirements to use the cuDNN implementation are:

    1. `activation` == `tanh`
    2. `recurrent_activation` == `sigmoid`
    3. `recurrent_dropout` == 0
    4. `unroll` is `False`
    5. `use_bias` is `True`
    6. Inputs, if use masking, are strictly right-padded.
    7. Eager execution is enabled in the outermost context.

    For example:

    >>> inputs = tf.random.normal([32, 10, 8])
    >>> lstm = tf.keras.layers.LSTM(4)
    >>> output = lstm(inputs)
    >>> print(output.shape)
    (32, 4)
    >>> lstm = tf.keras.layers.LSTM(4, return_sequences=True, return_state=True)
    >>> whole_seq_output, final_memory_state, final_carry_state = lstm(inputs)
    >>> print(whole_seq_output.shape)
    (32, 10, 4)
    >>> print(final_memory_state.shape)
    (32, 4)
    >>> print(final_carry_state.shape)
    (32, 4)

    Args:
      units: Positive integer, dimensionality of the output space.
      activation: Activation function to use.
        Default: hyperbolic tangent (`tanh`). If you pass `None`, no activation
        is applied (ie. "linear" activation: `a(x) = x`).
      recurrent_activation: Activation function to use for the recurrent step.
        Default: sigmoid (`sigmoid`). If you pass `None`, no activation is
        applied (ie. "linear" activation: `a(x) = x`).
      use_bias: Boolean (default `True`), whether the layer uses a bias vector.
      kernel_initializer: Initializer for the `kernel` weights matrix, used for
        the linear transformation of the inputs. Default: `glorot_uniform`.
      recurrent_initializer: Initializer for the `recurrent_kernel` weights
        matrix, used for the linear transformation of the recurrent state.
        Default: `orthogonal`.
      bias_initializer: Initializer for the bias vector. Default: `zeros`.
      unit_forget_bias: Boolean (default `True`). If True, add 1 to the bias of
        the forget gate at initialization. Setting it to true will also force
        `bias_initializer="zeros"`. This is recommended in [Jozefowicz et
            al.](http://www.jmlr.org/proceedings/papers/v37/jozefowicz15.pdf).
      kernel_regularizer: Regularizer function applied to the `kernel` weights
        matrix. Default: `None`.
      recurrent_regularizer: Regularizer function applied to the
        `recurrent_kernel` weights matrix. Default: `None`.
      bias_regularizer: Regularizer function applied to the bias vector.
        Default: `None`.
      activity_regularizer: Regularizer function applied to the output of the
        layer (its "activation"). Default: `None`.
      kernel_constraint: Constraint function applied to the `kernel` weights
        matrix. Default: `None`.
      recurrent_constraint: Constraint function applied to the
        `recurrent_kernel` weights matrix. Default: `None`.
      bias_constraint: Constraint function applied to the bias vector. Default:
        `None`.
      dropout: Float between 0 and 1. Fraction of the units to drop for the
        linear transformation of the inputs. Default: 0.
      recurrent_dropout: Float between 0 and 1. Fraction of the units to drop
        for the linear transformation of the recurrent state. Default: 0.
      return_sequences: Boolean. Whether to return the last output in the output
        sequence, or the full sequence. Default: `False`.
      return_state: Boolean. Whether to return the last state in addition to the
        output. Default: `False`.
      go_backwards: Boolean (default `False`). If True, process the input
        sequence backwards and return the reversed sequence.
      stateful: Boolean (default `False`). If True, the last state for each
      sample at index i in a batch will be used as initial state for the sample
        of index i in the following batch.
      time_major: The shape format of the `inputs` and `outputs` tensors.
        If True, the inputs and outputs will be in shape
        `[timesteps, batch, feature]`, whereas in the False case, it will be
        `[batch, timesteps, feature]`. Using `time_major = True` is a bit more
        efficient because it avoids transposes at the beginning and end of the
        RNN calculation. However, most TensorFlow data is batch-major, so by
        default this function accepts input and emits output in batch-major
        form.
      unroll: Boolean (default `False`). If True, the network will be unrolled,
        else a symbolic loop will be used. Unrolling can speed-up a RNN,
        although it tends to be more memory-intensive. Unrolling is only
        suitable for short sequences.

    Call arguments:
      inputs: A 3D tensor with shape `[batch, timesteps, feature]`.
      mask: Binary tensor of shape `[batch, timesteps]` indicating whether
        a given timestep should be masked (optional, defaults to `None`).
        An individual `True` entry indicates that the corresponding timestep
        should be utilized, while a `False` entry indicates that the
        corresponding timestep should be ignored.
      training: Python boolean indicating whether the layer should behave in
        training mode or in inference mode. This argument is passed to the cell
        when calling it. This is only relevant if `dropout` or
        `recurrent_dropout` is used (optional, defaults to `None`).
      initial_state: List of initial state tensors to be passed to the first
        call of the cell (optional, defaults to `None` which causes creation
        of zero-filled initial state tensors).
    """

    def __init__(
        self,
        units,
        activation="tanh",
        recurrent_activation="sigmoid",
        use_bias=True,
        kernel_initializer="glorot_uniform",
        recurrent_initializer="orthogonal",
        bias_initializer="zeros",
        unit_forget_bias=True,
        kernel_regularizer=None,
        recurrent_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        recurrent_constraint=None,
        bias_constraint=None,
        dropout=0.0,
        recurrent_dropout=0.0,
        return_sequences=False,
        return_state=False,
        go_backwards=False,
        stateful=False,
        time_major=False,
        unroll=False,
        **kwargs,
    ):
        # return_runtime is a flag for testing, which shows the real backend
        # implementation chosen by grappler in graph mode.
        self.return_runtime = kwargs.pop("return_runtime", False)
        implementation = kwargs.pop("implementation", 2)
        if implementation == 0:
            logging.warning(
                "`implementation=0` has been deprecated, "
                "and now defaults to `implementation=1`."
                "Please update your layer call."
            )
        if "enable_caching_device" in kwargs:
            cell_kwargs = {
                "enable_caching_device": kwargs.pop("enable_caching_device")
            }
        else:
            cell_kwargs = {}
        cell = LSTMCell(
            units,
            activation=activation,
            recurrent_activation=recurrent_activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            recurrent_initializer=recurrent_initializer,
            unit_forget_bias=unit_forget_bias,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            recurrent_regularizer=recurrent_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            recurrent_constraint=recurrent_constraint,
            bias_constraint=bias_constraint,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            implementation=implementation,
            dtype=kwargs.get("dtype"),
            trainable=kwargs.get("trainable", True),
            **cell_kwargs,
        )
        super().__init__(
            cell,
            return_sequences=return_sequences,
            return_state=return_state,
            go_backwards=go_backwards,
            stateful=stateful,
            time_major=time_major,
            unroll=unroll,
            **kwargs,
        )
        self.activity_regularizer = regularizers.get(activity_regularizer)
        self.input_spec = [InputSpec(ndim=3)]
        self.state_spec = [
            InputSpec(shape=(None, dim)) for dim in (self.units, self.units)
        ]
        self._could_use_gpu_kernel = (
            self.activation in (activations.tanh, tf.tanh)
            and self.recurrent_activation in (activations.sigmoid, tf.sigmoid)
            and recurrent_dropout == 0
            and not unroll
            and use_bias
            and tf.compat.v1.executing_eagerly_outside_functions()
        )
        if tf.config.list_logical_devices("GPU"):
            # Only show the message when there is GPU available, user will not
            # care about the cuDNN if there isn't any GPU.
            if self._could_use_gpu_kernel:
                logging.debug(gru_lstm_utils.CUDNN_AVAILABLE_MSG % self.name)
            else:
                logging.warning(
                    gru_lstm_utils.CUDNN_NOT_AVAILABLE_MSG % self.name
                )

        if gru_lstm_utils.use_new_gru_lstm_impl():
            self._defun_wrapper = gru_lstm_utils.DefunWrapper(
                time_major, go_backwards, "lstm"
            )

    def call(self, inputs, mask=None, training=None, initial_state=None):
        # The input should be dense, padded with zeros. If a ragged input is fed
        # into the layer, it is padded and the row lengths are used for masking.
        inputs, row_lengths = backend.convert_inputs_if_ragged(inputs)
        is_ragged_input = row_lengths is not None
        self._validate_args_if_ragged(is_ragged_input, mask)

        # LSTM does not support constants. Ignore it during process.
        inputs, initial_state, _ = self._process_inputs(
            inputs, initial_state, None
        )

        if isinstance(mask, list):
            mask = mask[0]

        input_shape = backend.int_shape(inputs)
        timesteps = input_shape[0] if self.time_major else input_shape[1]

        if not self._could_use_gpu_kernel:
            # Fall back to use the normal LSTM.
            kwargs = {"training": training}
            self._maybe_reset_cell_dropout_mask(self.cell)

            def step(inputs, states):
                return self.cell(inputs, states, **kwargs)

            last_output, outputs, states = backend.rnn(
                step,
                inputs,
                initial_state,
                constants=None,
                go_backwards=self.go_backwards,
                mask=mask,
                unroll=self.unroll,
                input_length=row_lengths
                if row_lengths is not None
                else timesteps,
                time_major=self.time_major,
                zero_output_for_mask=self.zero_output_for_mask,
                return_all_outputs=self.return_sequences,
            )
            runtime = gru_lstm_utils.runtime(gru_lstm_utils.RUNTIME_UNKNOWN)
        else:
            # Use the new defun approach for backend implementation swap.
            # Note that different implementations need to have same function
            # signature, eg, the tensor parameters need to have same shape and
            # dtypes. Since the cuDNN has an extra set of bias, those bias will
            # be passed to both normal and cuDNN implementations.
            self.reset_dropout_mask()
            dropout_mask = self.get_dropout_mask_for_cell(
                inputs, training, count=4
            )
            if dropout_mask is not None:
                inputs = inputs * dropout_mask[0]
            if gru_lstm_utils.use_new_gru_lstm_impl():
                lstm_kwargs = {
                    "inputs": inputs,
                    "init_h": gru_lstm_utils.read_variable_value(
                        initial_state[0]
                    ),
                    "init_c": gru_lstm_utils.read_variable_value(
                        initial_state[1]
                    ),
                    "kernel": gru_lstm_utils.read_variable_value(
                        self.cell.kernel
                    ),
                    "recurrent_kernel": gru_lstm_utils.read_variable_value(
                        self.cell.recurrent_kernel
                    ),
                    "bias": gru_lstm_utils.read_variable_value(self.cell.bias),
                    "mask": mask,
                    "time_major": self.time_major,
                    "go_backwards": self.go_backwards,
                    "sequence_lengths": row_lengths,
                    "zero_output_for_mask": self.zero_output_for_mask,
                }
                (
                    last_output,
                    outputs,
                    new_h,
                    new_c,
                    runtime,
                ) = self._defun_wrapper.defun_layer(**lstm_kwargs)
            else:
                gpu_lstm_kwargs = {
                    "inputs": inputs,
                    "init_h": gru_lstm_utils.read_variable_value(
                        initial_state[0]
                    ),
                    "init_c": gru_lstm_utils.read_variable_value(
                        initial_state[1]
                    ),
                    "kernel": gru_lstm_utils.read_variable_value(
                        self.cell.kernel
                    ),
                    "recurrent_kernel": gru_lstm_utils.read_variable_value(
                        self.cell.recurrent_kernel
                    ),
                    "bias": gru_lstm_utils.read_variable_value(self.cell.bias),
                    "mask": mask,
                    "time_major": self.time_major,
                    "go_backwards": self.go_backwards,
                    "sequence_lengths": row_lengths,
                    "return_sequences": self.return_sequences,
                }
                normal_lstm_kwargs = gpu_lstm_kwargs.copy()
                normal_lstm_kwargs.update(
                    {
                        "zero_output_for_mask": self.zero_output_for_mask,
                    }
                )

                if tf.executing_eagerly():
                    device_type = gru_lstm_utils.get_context_device_type()
                    can_use_gpu = (
                        # Either user specified GPU or unspecified but GPU is
                        # available.
                        (
                            device_type == gru_lstm_utils.GPU_DEVICE_NAME
                            or (
                                device_type is None
                                and tf.config.list_logical_devices("GPU")
                            )
                        )
                        and gru_lstm_utils.is_cudnn_supported_inputs(
                            mask, self.time_major, row_lengths
                        )
                    )
                    # Under eager context, check the device placement and prefer
                    # the GPU implementation when GPU is available.
                    if can_use_gpu:
                        last_output, outputs, new_h, new_c, runtime = gpu_lstm(
                            **gpu_lstm_kwargs
                        )
                    else:
                        (
                            last_output,
                            outputs,
                            new_h,
                            new_c,
                            runtime,
                        ) = standard_lstm(**normal_lstm_kwargs)
                else:
                    (
                        last_output,
                        outputs,
                        new_h,
                        new_c,
                        runtime,
                    ) = lstm_with_backend_selection(**normal_lstm_kwargs)

            states = [new_h, new_c]

        if self.stateful:
            updates = [
                tf.compat.v1.assign(
                    self_state, tf.cast(state, self_state.dtype)
                )
                for self_state, state in zip(self.states, states)
            ]
            self.add_update(updates)

        if self.return_sequences:
            output = backend.maybe_convert_to_ragged(
                is_ragged_input,
                outputs,
                row_lengths,
                go_backwards=self.go_backwards,
            )
        else:
            output = last_output

        if self.return_state:
            return [output] + list(states)
        elif self.return_runtime:
            return output, runtime
        else:
            return output

    @property
    def units(self):
        return self.cell.units

    @property
    def activation(self):
        return self.cell.activation

    @property
    def recurrent_activation(self):
        return self.cell.recurrent_activation

    @property
    def use_bias(self):
        return self.cell.use_bias

    @property
    def kernel_initializer(self):
        return self.cell.kernel_initializer

    @property
    def recurrent_initializer(self):
        return self.cell.recurrent_initializer

    @property
    def bias_initializer(self):
        return self.cell.bias_initializer

    @property
    def unit_forget_bias(self):
        return self.cell.unit_forget_bias

    @property
    def kernel_regularizer(self):
        return self.cell.kernel_regularizer

    @property
    def recurrent_regularizer(self):
        return self.cell.recurrent_regularizer

    @property
    def bias_regularizer(self):
        return self.cell.bias_regularizer

    @property
    def kernel_constraint(self):
        return self.cell.kernel_constraint

    @property
    def recurrent_constraint(self):
        return self.cell.recurrent_constraint

    @property
    def bias_constraint(self):
        return self.cell.bias_constraint

    @property
    def dropout(self):
        return self.cell.dropout

    @property
    def recurrent_dropout(self):
        return self.cell.recurrent_dropout

    @property
    def implementation(self):
        return self.cell.implementation

    def get_config(self):
        config = {
            "units": self.units,
            "activation": activations.serialize(self.activation),
            "recurrent_activation": activations.serialize(
                self.recurrent_activation
            ),
            "use_bias": self.use_bias,
            "kernel_initializer": initializers.serialize(
                self.kernel_initializer
            ),
            "recurrent_initializer": initializers.serialize(
                self.recurrent_initializer
            ),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "unit_forget_bias": self.unit_forget_bias,
            "kernel_regularizer": regularizers.serialize(
                self.kernel_regularizer
            ),
            "recurrent_regularizer": regularizers.serialize(
                self.recurrent_regularizer
            ),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
            "activity_regularizer": regularizers.serialize(
                self.activity_regularizer
            ),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "recurrent_constraint": constraints.serialize(
                self.recurrent_constraint
            ),
            "bias_constraint": constraints.serialize(self.bias_constraint),
            "dropout": self.dropout,
            "recurrent_dropout": self.recurrent_dropout,
            "implementation": self.implementation,
        }
        config.update(rnn_utils.config_for_enable_caching_device(self.cell))
        base_config = super().get_config()
        del base_config["cell"]
        return dict(list(base_config.items()) + list(config.items()))

    @classmethod
    def from_config(cls, config):
        if "implementation" in config and config["implementation"] == 0:
            config["implementation"] = 1
        return cls(**config)


def standard_lstm(
    inputs,
    init_h,
    init_c,
    kernel,
    recurrent_kernel,
    bias,
    mask,
    time_major,
    go_backwards,
    sequence_lengths,
    zero_output_for_mask,
    return_sequences,
):
    """LSTM with standard kernel implementation.

    This implementation can be run on all types for hardware.

    This implementation lifts out all the layer weights and make them function
    parameters. It has same number of tensor input params as the cuDNN
    counterpart. The RNN step logic has been simplified, eg dropout and mask is
    removed since cuDNN implementation does not support that.

    Note that the first half of the bias tensor should be ignored by this impl.
    The cuDNN impl need an extra set of input gate bias. In order to make the
    both function take same shape of parameter, that extra set of bias is also
    feed
    here.

    Args:
      inputs: input tensor of LSTM layer.
      init_h: initial state tensor for the cell output.
      init_c: initial state tensor for the cell hidden state.
      kernel: weights for cell kernel.
      recurrent_kernel: weights for cell recurrent kernel.
      bias: weights for cell kernel bias and recurrent bias. Only recurrent bias
        is used in this case.
      mask: Boolean tensor for mask out the steps within sequence.
        An individual `True` entry indicates that the corresponding timestep
        should be utilized, while a `False` entry indicates that the
        corresponding timestep should be ignored.
      time_major: boolean, whether the inputs are in the format of
        [time, batch, feature] or [batch, time, feature].
      go_backwards: Boolean (default False). If True, process the input sequence
        backwards and return the reversed sequence.
      sequence_lengths: The lengths of all sequences coming from a variable
        length input, such as ragged tensors. If the input has a fixed timestep
        size, this should be None.
      zero_output_for_mask: Boolean, whether to output zero for masked timestep.
      return_sequences: Boolean. If True, return the recurrent outputs for all
        timesteps in the sequence. If False, only return the output for the
        last timestep (which consumes less memory).

    Returns:
      last_output: output tensor for the last timestep, which has shape
        [batch, units].
      outputs:
        - If `return_sequences=True`: output tensor for all timesteps,
          which has shape [batch, time, units].
        - Else, a tensor equal to `last_output` with shape [batch, 1, units]
      state_0: the cell output, which has same shape as init_h.
      state_1: the cell hidden state, which has same shape as init_c.
      runtime: constant string tensor which indicate real runtime hardware. This
        value is for testing purpose and should be used by user.
    """
    input_shape = backend.int_shape(inputs)
    timesteps = input_shape[0] if time_major else input_shape[1]

    def step(cell_inputs, cell_states):
        """Step function that will be used by Keras RNN backend."""
        h_tm1 = cell_states[0]  # previous memory state
        c_tm1 = cell_states[1]  # previous carry state

        z = backend.dot(cell_inputs, kernel)
        z += backend.dot(h_tm1, recurrent_kernel)
        z = backend.bias_add(z, bias)

        z0, z1, z2, z3 = tf.split(z, 4, axis=1)

        i = tf.sigmoid(z0)
        f = tf.sigmoid(z1)
        c = f * c_tm1 + i * tf.tanh(z2)
        o = tf.sigmoid(z3)

        h = o * tf.tanh(c)
        return h, [h, c]

    last_output, outputs, new_states = backend.rnn(
        step,
        inputs,
        [init_h, init_c],
        constants=None,
        unroll=False,
        time_major=time_major,
        mask=mask,
        go_backwards=go_backwards,
        input_length=(
            sequence_lengths if sequence_lengths is not None else timesteps
        ),
        zero_output_for_mask=zero_output_for_mask,
        return_all_outputs=return_sequences,
    )
    return (
        last_output,
        outputs,
        new_states[0],
        new_states[1],
        gru_lstm_utils.runtime(gru_lstm_utils.RUNTIME_CPU),
    )


def gpu_lstm(
    inputs,
    init_h,
    init_c,
    kernel,
    recurrent_kernel,
    bias,
    mask,
    time_major,
    go_backwards,
    sequence_lengths,
    return_sequences,
):
    """LSTM with either cuDNN or ROCm implementation which is only available for
    GPU.

    Note that currently only right padded data is supported, or the result will
    be polluted by the unmasked data which should be filtered.

    Args:
      inputs: Input tensor of LSTM layer.
      init_h: Initial state tensor for the cell output.
      init_c: Initial state tensor for the cell hidden state.
      kernel: Weights for cell kernel.
      recurrent_kernel: Weights for cell recurrent kernel.
      bias: Weights for cell kernel bias and recurrent bias. Only recurrent bias
        is used in this case.
      mask: Boolean tensor for mask out the steps within sequence. An individual
        `True` entry indicates that the corresponding timestep should be
        utilized, while a `False` entry indicates that the corresponding
        timestep should be ignored.
      time_major: Boolean, whether the inputs are in the format of [time, batch,
        feature] or [batch, time, feature].
      go_backwards: Boolean (default False). If True, process the input sequence
        backwards and return the reversed sequence.
      sequence_lengths: The lengths of all sequences coming from a variable
        length input, such as ragged tensors. If the input has a fixed timestep
        size, this should be None.
      return_sequences: Boolean. If True, return the recurrent outputs for all
        timesteps in the sequence. If False, only return the output for the
        last timestep, matching the CPU function output format.

    Returns:
      last_output: Output tensor for the last timestep, which has shape
        [batch, units].
      outputs:
        - If `return_sequences=True`: output tensor for all timesteps,
          which has shape [batch, time, units].
        - Else, a tensor equal to `last_output` with shape [batch, 1, units]
      state_0: The cell output, which has same shape as init_h.
      state_1: The cell hidden state, which has same shape as init_c.
      runtime: Constant string tensor which indicate real runtime hardware. This
        value is for testing purpose and should not be used by user.
    """
    if mask is not None:
        sequence_lengths = gru_lstm_utils.calculate_sequence_by_mask(
            mask, time_major
        )

    if not time_major and sequence_lengths is None:
        inputs = tf.transpose(inputs, perm=(1, 0, 2))
        seq_axis, batch_axis = (0, 1)
    else:
        seq_axis, batch_axis = (0, 1) if time_major else (1, 0)
    # For init_h and init_c, cuDNN expects one more dim of num_layers before or
    # after batch dim for time major or batch major inputs respectively
    init_h = tf.expand_dims(init_h, axis=seq_axis)
    init_c = tf.expand_dims(init_c, axis=seq_axis)

    weights = tf.split(kernel, 4, axis=1)
    weights += tf.split(recurrent_kernel, 4, axis=1)
    # cuDNN has an extra set of bias for inputs, we disable them (setting to 0),
    # so that mathematically it is same as the canonical LSTM implementation.
    full_bias = tf.concat((tf.zeros_like(bias), bias), 0)

    if tf.sysconfig.get_build_info()["is_rocm_build"]:
        # ROCm MIOpen's weight sequence for LSTM is different from both
        # canonical and Cudnn format
        # MIOpen: [i, f, o, c] Cudnn/Canonical: [i, f, c, o]
        # i is input gate weights.
        # f is forget gate weights.
        # o is output gate weights.
        # c is cell gate weights.
        weights = [weights[x] for x in (0, 1, 3, 2, 4, 5, 7, 6)]
        # full_bias is a tensor of shape (8*n,)
        full_bias = tf.split(full_bias, 8, axis=0)
        full_bias = [full_bias[x] for x in (0, 1, 3, 2, 4, 5, 7, 6)]

    params = gru_lstm_utils.canonical_to_params(
        weights=weights,
        biases=tf.split(full_bias, 8),
        shape=tf.constant([-1]),
        transpose_weights=True,
    )

    if sequence_lengths is not None:
        if go_backwards:
            # Three reversals are required. E.g.,
            # normal input = [1, 2, 3, 0, 0]  # where 0 need to be masked
            # reversed_input_to_cudnn = [3, 2, 1, 0, 0]
            # output_from_cudnn = [6, 5, 4, 0, 0]
            # expected_output = [0, 0, 6, 5 ,4]
            inputs = tf.reverse_sequence(
                inputs,
                sequence_lengths,
                seq_axis=seq_axis,
                batch_axis=batch_axis,
            )
        outputs, h, c, _, _ = tf.raw_ops.CudnnRNNV3(
            input=inputs,
            input_h=init_h,
            input_c=init_c,
            params=params,
            is_training=True,
            rnn_mode="lstm",
            sequence_lengths=sequence_lengths,
            time_major=time_major,
        )
        if go_backwards:
            outputs = tf.reverse_sequence(
                outputs,
                sequence_lengths,
                seq_axis=seq_axis,
                batch_axis=batch_axis,
            )
            outputs = tf.reverse(outputs, axis=[seq_axis])
    else:
        # # Fill the array with shape [batch] with value of max timesteps.
        # sequence_length = array_ops.fill([array_ops.shape(inputs)[1]],
        #                                  array_ops.shape(inputs)[0])
        if go_backwards:
            # Reverse axis 0 since the input is already convert to time major.
            inputs = tf.reverse(inputs, axis=[0])
        outputs, h, c, _ = tf.raw_ops.CudnnRNN(
            input=inputs,
            input_h=init_h,
            input_c=init_c,
            params=params,
            is_training=True,
            rnn_mode="lstm",
        )

    last_output = outputs[-1]
    if not time_major and sequence_lengths is None and return_sequences:
        outputs = tf.transpose(outputs, perm=[1, 0, 2])
    h = tf.squeeze(h, axis=seq_axis)
    c = tf.squeeze(c, axis=seq_axis)

    # In the case of variable length input, the cudnn kernel will fill zeros for
    # the output, whereas the default keras behavior is to bring over the
    # previous output for t-1, so that in the return_sequence=False case, user
    # can quickly get the final effect output instead just 0s at the last
    # timestep.  In order to mimic the default keras behavior, we copy the final
    # h state as the last_output, since it is numerically same as the output.
    if sequence_lengths is not None:
        last_output = h

    # Match CPU return format
    if not return_sequences:
        outputs = tf.expand_dims(last_output, axis=0 if time_major else 1)

    return (
        last_output,
        outputs,
        h,
        c,
        gru_lstm_utils.runtime(gru_lstm_utils.RUNTIME_GPU),
    )


def lstm_with_backend_selection(
    inputs,
    init_h,
    init_c,
    kernel,
    recurrent_kernel,
    bias,
    mask,
    time_major,
    go_backwards,
    sequence_lengths,
    zero_output_for_mask,
    return_sequences,
):
    """Call the LSTM with optimized backend kernel selection.

    Under the hood, this function will create two TF function, one with the most
    generic kernel and can run on all device condition, and the second one with
    cuDNN specific kernel, which can only run on GPU.

    The first function will be called with normal_lstm_params, while the second
    function is not called, but only registered in the graph. The Grappler will
    do the proper graph rewrite and swap the optimized TF function based on the
    device placement.

    Args:
      inputs: Input tensor of LSTM layer.
      init_h: Initial state tensor for the cell output.
      init_c: Initial state tensor for the cell hidden state.
      kernel: Weights for cell kernel.
      recurrent_kernel: Weights for cell recurrent kernel.
      bias: Weights for cell kernel bias and recurrent bias. Only recurrent bias
        is used in this case.
      mask: Boolean tensor for mask out the steps within sequence.
        An individual `True` entry indicates that the corresponding timestep
        should be utilized, while a `False` entry indicates that the
        corresponding timestep should be ignored.
      time_major: Boolean, whether the inputs are in the format of
        [time, batch, feature] or [batch, time, feature].
      go_backwards: Boolean (default False). If True, process the input sequence
        backwards and return the reversed sequence.
      sequence_lengths: The lengths of all sequences coming from a variable
        length input, such as ragged tensors. If the input has a fixed timestep
        size, this should be None.
      zero_output_for_mask: Boolean, whether to output zero for masked timestep.
      return_sequences: Boolean. If True, return the recurrent outputs for all
        timesteps in the sequence. If False, only return the output for the
        last timestep (which consumes less memory).

    Returns:
      List of output tensors, same as standard_lstm.
    """
    params = {
        "inputs": inputs,
        "init_h": init_h,
        "init_c": init_c,
        "kernel": kernel,
        "recurrent_kernel": recurrent_kernel,
        "bias": bias,
        "mask": mask,
        "time_major": time_major,
        "go_backwards": go_backwards,
        "sequence_lengths": sequence_lengths,
        "zero_output_for_mask": zero_output_for_mask,
        "return_sequences": return_sequences,
    }

    def gpu_lstm_with_fallback(
        inputs,
        init_h,
        init_c,
        kernel,
        recurrent_kernel,
        bias,
        mask,
        time_major,
        go_backwards,
        sequence_lengths,
        zero_output_for_mask,
        return_sequences,
    ):
        """Use cuDNN kernel when mask is none or strictly right padded."""

        def cudnn_lstm_fn():
            return gpu_lstm(
                inputs=inputs,
                init_h=init_h,
                init_c=init_c,
                kernel=kernel,
                recurrent_kernel=recurrent_kernel,
                bias=bias,
                mask=mask,
                time_major=time_major,
                go_backwards=go_backwards,
                sequence_lengths=sequence_lengths,
                return_sequences=return_sequences,
            )

        def stardard_lstm_fn():
            return standard_lstm(
                inputs=inputs,
                init_h=init_h,
                init_c=init_c,
                kernel=kernel,
                recurrent_kernel=recurrent_kernel,
                bias=bias,
                mask=mask,
                time_major=time_major,
                go_backwards=go_backwards,
                sequence_lengths=sequence_lengths,
                zero_output_for_mask=zero_output_for_mask,
                return_sequences=return_sequences,
            )

        return tf.__internal__.smart_cond.smart_cond(
            gru_lstm_utils.is_cudnn_supported_inputs(
                mask, time_major, sequence_lengths
            ),
            true_fn=cudnn_lstm_fn,
            false_fn=stardard_lstm_fn,
        )

    if gru_lstm_utils.use_new_gru_lstm_impl():
        # Chooses the implementation dynamically based on the running device.
        (
            last_output,
            outputs,
            new_h,
            new_c,
            runtime,
        ) = tf.__internal__.execute_fn_for_device(
            {
                gru_lstm_utils.CPU_DEVICE_NAME: lambda: standard_lstm(**params),
                gru_lstm_utils.GPU_DEVICE_NAME: lambda: gpu_lstm_with_fallback(
                    **params
                ),
            },
            lambda: standard_lstm(**params),
        )
    else:
        # Each time a `tf.function` is called, we will give it a unique
        # identifiable API name, so that Grappler won't get confused when it
        # sees multiple LSTM layers added into same graph, and it will be able
        # to pair up the different implementations across them.
        api_name = "lstm_" + str(uuid.uuid4())
        supportive_attribute = {
            "time_major": time_major,
            "go_backwards": go_backwards,
        }
        defun_standard_lstm = gru_lstm_utils.generate_defun_backend(
            api_name,
            gru_lstm_utils.CPU_DEVICE_NAME,
            standard_lstm,
            supportive_attribute,
        )
        defun_gpu_lstm = gru_lstm_utils.generate_defun_backend(
            api_name,
            gru_lstm_utils.GPU_DEVICE_NAME,
            gpu_lstm_with_fallback,
            supportive_attribute,
        )

        # Call the normal LSTM impl and register the cuDNN impl function. The
        # grappler will kick in during session execution to optimize the graph.
        last_output, outputs, new_h, new_c, runtime = defun_standard_lstm(
            **params
        )
        gru_lstm_utils.function_register(defun_gpu_lstm, **params)

    return last_output, outputs, new_h, new_c, runtime



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Keras SavedModel deserialization."""

import re
import types
import warnings

import tensorflow.compat.v1.logging as logging
import tensorflow.compat.v2 as tf
from google.protobuf import message

from keras import backend
from keras import regularizers
from keras.engine import input_spec
from keras.optimizers.legacy import optimizer_v2
from keras.protobuf import saved_metadata_pb2
from keras.protobuf import versions_pb2
from keras.saving import object_registration
from keras.saving.legacy import saving_utils
from keras.saving.legacy import serialization
from keras.saving.legacy.saved_model import constants
from keras.saving.legacy.saved_model import json_utils
from keras.saving.legacy.saved_model import utils
from keras.saving.legacy.saved_model.serialized_attributes import (
    CommonEndpoints,
)
from keras.utils import layer_utils
from keras.utils import metrics_utils
from keras.utils import tf_inspect
from keras.utils.generic_utils import LazyLoader

# To avoid circular dependencies between keras/engine and keras/saving,
# code in keras/saving must delay imports.

# TODO(b/134426265): Switch back to single-quotes to match the rest of the file
# once the issue with copybara is fixed.

models_lib = LazyLoader("models_lib", globals(), "keras.models")
base_layer = LazyLoader("base_layer", globals(), "keras.engine.base_layer")
layers_module = LazyLoader("layers_module", globals(), "keras.layers")
input_layer = LazyLoader("input_layer", globals(), "keras.engine.input_layer")
functional_lib = LazyLoader(
    "functional_lib", globals(), "keras.engine.functional"
)
training_lib = LazyLoader("training_lib", globals(), "keras.engine.training")
training_lib_v1 = LazyLoader(
    "training_lib_v1", globals(), "keras.engine.training_v1"
)
metrics = LazyLoader("metrics", globals(), "keras.metrics")
base_rnn = LazyLoader("base_rnn", globals(), "keras.layers.rnn.base_rnn")


PUBLIC_ATTRIBUTES = CommonEndpoints.all_functions.union(
    CommonEndpoints.all_checkpointable_objects
)
PUBLIC_ATTRIBUTES.add(constants.KERAS_ATTR)


def load(path, compile=True, options=None):
    """Loads Keras objects from a SavedModel.

    Any Keras layer or model saved to the SavedModel will be loaded back
    as Keras objects. Other objects are loaded as regular trackable objects
    (same as `tf.saved_model.load`).

    Currently, Keras saving/loading only retains the Keras object's weights,
    losses, and call function.

    The loaded model can be re-compiled, but the original optimizer, compiled
    loss functions, and metrics are not retained. This is temporary, and
    `model.save` will soon be able to serialize compiled models.

    Args:
      path: Path to SavedModel.
      compile: If true, compile the model after loading it.
      options: Optional `tf.saved_model.LoadOptions` object that specifies
        options for loading from SavedModel.

    Returns:
      Object loaded from SavedModel.
    """
    # TODO(kathywu): Add saving/loading of optimizer, compiled losses and
    # metrics.
    # TODO(kathywu): Add code to load from objects that contain all endpoints

    # Look for metadata file or parse the SavedModel
    metadata = saved_metadata_pb2.SavedMetadata()
    meta_graph_def = tf.__internal__.saved_model.parse_saved_model(
        path
    ).meta_graphs[0]
    object_graph_def = meta_graph_def.object_graph_def
    path_to_metadata_pb = tf.io.gfile.join(path, constants.SAVED_METADATA_PATH)
    if tf.compat.v1.gfile.Exists(path_to_metadata_pb):
        try:
            with tf.io.gfile.GFile(path_to_metadata_pb, "rb") as f:
                file_content = f.read()
            metadata.ParseFromString(file_content)
        except message.DecodeError as e:
            raise IOError(
                f"Cannot parse keras metadata at path {path_to_metadata_pb}: "
                f"Received error: {e}"
            )
    else:
        logging.warning(
            "SavedModel saved prior to TF 2.5 detected when loading "
            "Keras model. Please ensure that you are saving the model "
            "with model.save() or tf.keras.models.save_model(), *NOT* "
            "tf.saved_model.save(). To confirm, there should be a file "
            'named "keras_metadata.pb" in the SavedModel directory.'
        )
        _read_legacy_metadata(object_graph_def, metadata, path)

    if not metadata.nodes:
        # When there are no Keras objects, return the results from the core
        # loader
        return tf.saved_model.load(path, options=options)

    metadata = _update_to_current_version(metadata)
    # Recreate layers and metrics using the info stored in the metadata.
    keras_loader = KerasObjectLoader(metadata, object_graph_def)
    keras_loader.load_layers(compile=compile)

    # Generate a dictionary of all loaded nodes.
    nodes_to_load = {"root": None}
    for node_id, loaded_node in keras_loader.loaded_nodes.items():
        nodes_to_load[keras_loader.get_path(node_id)] = loaded_node
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message="Trying to load ShardedVariables"
        )
        loaded = tf.__internal__.saved_model.load_partial(
            path, nodes_to_load, options=options
        )

    # Finalize the loaded layers and remove the extra tracked dependencies.
    keras_loader.finalize_objects()
    keras_loader.del_tracking()

    model = loaded["root"]

    if isinstance(model, training_lib.Model) and compile:
        # TODO(kathywu): Use compiled objects from SavedModel, instead of
        # creating new objects from the training config.
        training_config = model._serialized_attributes["metadata"].get(
            "training_config", None
        )
        if training_config is not None:
            model.compile(
                **saving_utils.compile_args_from_training_config(
                    training_config
                ),
                from_serialized=True,
            )
            saving_utils.try_build_compiled_arguments(model)
            if isinstance(model.optimizer, optimizer_v2.OptimizerV2):
                if model.optimizer.get_slot_names():
                    logging.warning(
                        "Your optimizer uses slots. "
                        "Slots cannot be restored from saved_model, "
                        "as a result, your model is starting with  "
                        "a new initialized optimizer."
                    )
        else:
            logging.warning(
                "No training configuration found in save file, so the "
                "model was *not* compiled. Compile it manually."
            )

    # Force variables and resources to initialize.
    if not tf.executing_eagerly():
        sess = backend.get_session()  # Variables are initialized by this call.
        sess.run(
            tf.compat.v1.get_collection(
                tf.compat.v1.GraphKeys.TABLE_INITIALIZERS
            )
        )

    return model


def _update_to_current_version(metadata):
    """Applies version updates to the metadata proto for backwards compat."""
    for node in metadata.nodes:
        if node.version.producer == 1 and node.identifier in [
            constants.MODEL_IDENTIFIER,
            constants.SEQUENTIAL_IDENTIFIER,
            constants.NETWORK_IDENTIFIER,
        ]:
            node_metadata = json_utils.decode(node.metadata)
            save_spec = node_metadata.get("save_spec")

            if save_spec is not None:
                node_metadata["full_save_spec"] = ([save_spec], {})
                node.metadata = json_utils.Encoder().encode(node_metadata)
    return metadata


def _read_legacy_metadata(object_graph_def, metadata, path):
    """Builds a KerasMetadata proto from the SavedModel ObjectGraphDef."""
    # Older SavedModels store the metadata directly in the proto instead of the
    # separate pb file.
    node_paths = _generate_object_paths(object_graph_def)
    for node_id, proto in enumerate(object_graph_def.nodes):
        if (
            proto.WhichOneof("kind") == "user_object"
            and proto.user_object.identifier
            in constants.KERAS_OBJECT_IDENTIFIERS
        ):
            if not proto.user_object.metadata:
                raise ValueError(
                    "Unable to create a Keras model from SavedModel at "
                    f"{path}. This SavedModel was exported with "
                    "`tf.saved_model.save`, and lacks the Keras metadata file. "
                    "Please save your Keras model by calling `model.save` "
                    "or `tf.keras.models.save_model`. Note that "
                    "you can still load this SavedModel with "
                    "`tf.saved_model.load`."
                )
            metadata.nodes.add(
                node_id=node_id,
                node_path=node_paths[node_id],
                version=versions_pb2.VersionDef(
                    producer=1, min_consumer=1, bad_consumers=[]
                ),
                identifier=proto.user_object.identifier,
                metadata=proto.user_object.metadata,
            )


def _generate_object_paths(object_graph_def):
    """Traverses through an ObjectGraphDef and builds a map of all node
    paths."""
    paths = {0: "root"}
    nodes_to_visit = [0]

    while nodes_to_visit:
        current_node = nodes_to_visit.pop()
        current_path = paths[current_node]
        for reference in object_graph_def.nodes[current_node].children:
            if reference.node_id in paths:
                continue
            paths[reference.node_id] = f"{current_path}.{reference.local_name}"
            nodes_to_visit.append(reference.node_id)

    return paths


def _is_graph_network(layer):
    """Determines whether the layer is a graph network."""

    if isinstance(layer, RevivedNetwork):
        return False
    elif isinstance(layer, functional_lib.Functional):
        return layer._is_graph_network or isinstance(
            layer, models_lib.Sequential
        )
    return False


class KerasObjectLoader:
    """Loader that recreates Keras objects (e.g.

    layers, models).

    Layers and models are revived from either the config or SavedModel following
    these rules:
    1. If object is a graph network (i.e. Sequential or Functional) then it will
       be initialized using the structure from the config only after the
       children layers have been created. Graph networks must be initialized
       with inputs and outputs, so all child layers must be created beforehand.
    2. If object's config exists and the class can be found, then revive from
       config.
    3. Object may have already been created if its parent was revived from
       config. In this case, do nothing.
    4. If nothing of the above applies, compose the various artifacts from the
       SavedModel to create a subclassed layer or model. At this time, custom
       metrics are not supported.

    """

    def __init__(self, metadata, object_graph_def):
        self._metadata = {x.node_id: x for x in metadata.nodes}
        self._proto = object_graph_def

        self._node_paths = {
            node_data.node_id: node_data.node_path
            for node_data in metadata.nodes
        }
        self.loaded_nodes = {}  # Maps node path -> loaded node

        # Store all node ids that have already been traversed when tracking
        # nodes that were recreated from the config.
        self._traversed_nodes_from_config = set()

        # Maps model id -> (blank model obj, list of child layer or their node
        # ids) This tracks all layers in functional and sequential models. These
        # models are only reconstructed after all of their child layers have
        # been created.
        self.model_layer_dependencies = {}
        self._models_to_reconstruct = []

    def del_tracking(self):
        """Removes tracked references that are only used when loading the
        model."""
        # Now that the node object has been fully loaded, and the checkpoint has
        # been restored, the object no longer needs to track objects added from
        # SerializedAttributes. (Note that saving a training checkpoint still
        # functions correctly, because layers and variables are tracked
        # separately by the Layer object.)
        # TODO(kathywu): Instead of outright deleting these nodes (which would
        # make restoring from a different checkpoint tricky), mark them as extra
        # dependencies that are OK to overwrite.
        for node in self.loaded_nodes.values():
            node = node[0]
            if not isinstance(node, base_layer.Layer):
                # Loaded nodes can contain other trackable objects created when
                # loading layers from the config, such as variables.
                continue
            for name in PUBLIC_ATTRIBUTES:
                node._delete_tracking(name)

            if isinstance(node, functional_lib.Functional):
                # Delete the temporary layer dependencies, which were used to
                # restore the checkpointed values. When the model is live, the
                # user can delete or add layers to the model at any time, so
                # these layer dependencies may be obsolete.
                dependencies = list(node._self_unconditional_dependency_names)
                for name in dependencies:
                    if (
                        re.match(r"^layer(_with_weights)?-[\d+]", name)
                        is not None
                    ):
                        node._delete_tracking(name)

    def _add_children_recreated_from_config(self, obj, proto, node_id):
        """Recursively records objects recreated from config."""

        if node_id in self._traversed_nodes_from_config:
            return

        parent_path = self._node_paths[node_id]
        self._traversed_nodes_from_config.add(node_id)
        obj._maybe_initialize_trackable()
        if isinstance(obj, base_layer.Layer) and not obj.built:
            metadata = json_utils.decode(self._metadata[node_id].metadata)
            self._try_build_layer(
                obj, node_id, metadata.get("build_input_shape")
            )

        # Create list of all possible children
        children = []
        # Look for direct children
        for reference in proto.children:
            obj_child = obj._lookup_dependency(reference.local_name)
            children.append(
                (obj_child, reference.node_id, reference.local_name)
            )

        # Add metrics that may have been added to the layer._metrics list.
        # This is stored in the SavedModel as layer.keras_api.layer_metrics in
        # SavedModels created after Tf 2.2.
        metric_list_node_id = self._search_for_child_node(
            node_id, [constants.KERAS_ATTR, "layer_metrics"]
        )
        if metric_list_node_id is not None and hasattr(obj, "_metrics"):
            obj_metrics = {m.name: m for m in obj._metrics}
            for reference in self._proto.nodes[metric_list_node_id].children:
                metric = obj_metrics.get(reference.local_name)
                if metric is not None:
                    metric_path = "{}.layer_metrics.{}".format(
                        constants.KERAS_ATTR, reference.local_name
                    )
                    children.append((metric, reference.node_id, metric_path))

        for obj_child, child_id, child_name in children:
            child_proto = self._proto.nodes[child_id]

            if not isinstance(obj_child, tf.__internal__.tracking.Trackable):
                continue
            if (
                child_proto.user_object.identifier
                in tf.__internal__.saved_model.load.registered_identifiers()
            ):
                setter = tf.__internal__.saved_model.load.get_setter(
                    child_proto.user_object
                )
            elif (
                obj_child._object_identifier
                in constants.KERAS_OBJECT_IDENTIFIERS
            ):
                setter = _revive_setter
            else:
                setter = setattr

            if child_id in self.loaded_nodes:
                if self.loaded_nodes[child_id][0] is not obj_child:
                    # This means that the same trackable object is referenced by
                    # two different objects that were recreated from the config.
                    logging.warning(
                        "Looks like there is an object (perhaps variable or "
                        "layer) that is shared between different "
                        "layers/models. This may cause issues when restoring "
                        "the variable values. Object: {}".format(obj_child)
                    )
                continue

            # Overwrite variable names with the ones saved in the SavedModel.
            if (
                child_proto.WhichOneof("kind") == "variable"
                and child_proto.variable.name
            ):
                obj_child._handle_name = child_proto.variable.name + ":0"

            if isinstance(
                obj_child, tf.__internal__.tracking.TrackableDataStructure
            ):
                setter = lambda *args: None

            child_path = f"{parent_path}.{child_name}"
            self._node_paths[child_id] = child_path
            self._add_children_recreated_from_config(
                obj_child, child_proto, child_id
            )
            self.loaded_nodes[child_id] = obj_child, setter

    def load_layers(self, compile=True):
        """Load all layer nodes from the metadata."""
        # Load metrics after models and layers, since it's likely that models
        # and layers will create the metric when initialized (this avoids
        # wasting time by creating objects multiple times).
        metric_list = []
        for node_metadata in self._metadata.values():
            if node_metadata.identifier == constants.METRIC_IDENTIFIER:
                metric_list.append(node_metadata)
                continue

            self.loaded_nodes[node_metadata.node_id] = self._load_layer(
                node_metadata.node_id,
                node_metadata.identifier,
                node_metadata.metadata,
            )

        for node_metadata in metric_list:
            try:
                self.loaded_nodes[node_metadata.node_id] = self._load_layer(
                    node_metadata.node_id,
                    node_metadata.identifier,
                    node_metadata.metadata,
                )
            except ValueError as e:
                # Metrics are only needed when the model is compiled later. We
                # ignore errors when trying to load custom metrics when
                # `compile=False` until custom metrics are serialized properly
                # (b/135550038).
                if compile:
                    raise e
                logging.warning(
                    "Unable to restore custom metric. Please ensure that "
                    "the layer implements `get_config` and `from_config` "
                    "when saving. In addition, please use the "
                    "`custom_objects` arg when calling `load_model()`."
                )

    def _load_layer(self, node_id, identifier, metadata):
        """Load a single layer from a SavedUserObject proto."""
        metadata = json_utils.decode(metadata)

        # If node was already created
        if node_id in self.loaded_nodes:
            node, setter = self.loaded_nodes[node_id]

            # Revive setter requires the object to have a
            # `_serialized_attributes` property. Add it here.
            _maybe_add_serialized_attributes(node, metadata)

            config = metadata.get("config")
            if _is_graph_network(node) and serialization.validate_config(
                config
            ):
                child_nodes = self._get_child_layer_node_ids(node_id)
                self.model_layer_dependencies[node_id] = (node, child_nodes)
                if not child_nodes:
                    self._models_to_reconstruct.append(node_id)
            return node, setter

        # Detect whether this object can be revived from the config. If not,
        # then revive from the SavedModel instead.
        obj, setter = self._revive_from_config(identifier, metadata, node_id)
        if obj is None:
            obj, setter = revive_custom_object(identifier, metadata)

        # Add an attribute that stores the extra functions/objects saved in the
        # SavedModel. Most of these functions/objects are ignored, but some are
        # used later in the loading process (e.g. the list of regularization
        # losses, or the training config of compiled models).
        _maybe_add_serialized_attributes(obj, metadata)
        return obj, setter

    def _revive_from_config(self, identifier, metadata, node_id):
        """Revives a layer/model from config, or returns None."""
        if identifier == constants.METRIC_IDENTIFIER:
            obj = self._revive_metric_from_config(metadata)
        else:
            obj = self._revive_graph_network(
                identifier, metadata, node_id
            ) or self._revive_layer_or_model_from_config(metadata, node_id)

        if obj is None:
            return None, None

        setter = self._config_node_setter(_revive_setter)
        self._add_children_recreated_from_config(
            obj, self._proto.nodes[node_id], node_id
        )
        return obj, setter

    def _revive_graph_network(self, identifier, metadata, node_id):
        """Revives a graph network from config."""
        # Determine whether the metadata contains information for reviving a
        # functional or Sequential model.
        config = metadata.get("config")
        if not serialization.validate_config(config):
            return None

        class_name = tf.compat.as_str(metadata["class_name"])
        if object_registration.get_registered_object(class_name) is not None:
            return None
        model_is_functional_or_sequential = (
            metadata.get("is_graph_network", False)
            or class_name == "Sequential"
            or class_name == "Functional"
        )
        if not model_is_functional_or_sequential:
            return None

        # Revive functional and sequential models as blank model objects for now
        # ( must be initialized to enable setattr tracking and attribute
        # caching).  Reconstruction of the network is deferred until all of the
        # model's layers have been revived.
        if class_name == "Sequential":
            model = models_lib.Sequential(name=config["name"])
        # The model is a custom Sequential model.
        elif identifier == constants.SEQUENTIAL_IDENTIFIER:
            # Uses the custom class name, since the config does not have one.
            model = models_lib.Sequential(name=class_name)
        else:
            model = models_lib.Functional(
                inputs=[], outputs=[], name=config["name"]
            )

        # Record this model and its layers. This will later be used to
        # reconstruct the model.
        layers = self._get_child_layer_node_ids(node_id)
        self.model_layer_dependencies[node_id] = (model, layers)
        if not layers:
            self._models_to_reconstruct.append(node_id)
        return model

    def _revive_layer_or_model_from_config(self, metadata, node_id):
        """Revives a layer/custom model from config; returns None if
        infeasible."""
        # Check that the following requirements are met for reviving from
        # config:
        #    1. Object can be deserialized from config.
        #    2. If the object needs to be built, then the build input shape can
        #       be found.
        class_name = metadata.get("class_name")
        config = metadata.get("config")
        shared_object_id = metadata.get("shared_object_id")
        must_restore_from_config = metadata.get("must_restore_from_config")
        if not serialization.validate_config(config):
            return None

        try:
            try:
                obj = layers_module.deserialize(
                    serialization.serialize_keras_class_and_config(
                        class_name, config, shared_object_id=shared_object_id
                    )
                )
            except (TypeError, KeyError) as e:
                # A name conflict has occurred. The `class_name` is in the Keras
                # native framework; however, the value in the framework is
                # different from the user's class definition which confuses the
                # KerasObjectLoader.
                builtin_layer = layers_module.get_builtin_layer(class_name)
                if builtin_layer:
                    raise RuntimeError(
                        f"Unable to restore object of class '{class_name}'. "
                        "One of several possible causes could be "
                        "a missing custom object. "
                        "Decorate your custom object with "
                        "`@keras.utils.register_keras_serializable` and "
                        "include that file in your program, "
                        "or pass your class in a "
                        "`keras.utils.CustomObjectScope` "
                        "that wraps this load call. "
                        f"\n\nException: {e}"
                    ) from e
                else:
                    raise
        except Exception as e:
            if must_restore_from_config:
                raise e
            else:
                return None

        # Use the dtype, name, and trainable status. Often times these are not
        # specified in custom configs, so retrieve their values from the
        # metadata.

        obj._name = metadata["name"]
        if metadata.get("trainable") is not None:
            obj.trainable = metadata["trainable"]
        if metadata.get("dtype") is not None:
            obj._set_dtype_policy(metadata["dtype"])
        if metadata.get("stateful") is not None:
            obj.stateful = metadata["stateful"]
        if metadata.get("autocast") is not None:
            obj._autocast = metadata["autocast"]
        # Restore model save spec for subclassed models. (layers do not store a
        # SaveSpec)
        if isinstance(obj, training_lib.Model):
            full_save_spec = metadata.get("full_save_spec")
            if full_save_spec is not None:
                args_spec, kwargs_spec = full_save_spec
                inputs_spec = args_spec.pop(0)
                obj._set_save_spec(inputs_spec, args_spec, kwargs_spec)

        build_input_shape = metadata.get("build_input_shape")
        built = self._try_build_layer(obj, node_id, build_input_shape)

        if not built:
            # If the layer cannot be built, revive a custom layer instead.
            return None
        return obj

    def _revive_metric_from_config(self, metadata):
        """Revives a metric object using the config saved in the metadata."""
        class_name = tf.compat.as_str(metadata["class_name"])
        config = metadata.get("config")

        if not serialization.validate_config(config):
            return None

        try:
            obj = metrics.deserialize(
                serialization.serialize_keras_class_and_config(
                    class_name, config
                )
            )
        except ValueError:
            return None

        build_input_shape = metadata.get("build_input_shape")
        if build_input_shape is not None and hasattr(obj, "_build"):
            obj._build(build_input_shape)

        return obj

    def _try_build_layer(self, obj, node_id, build_input_shape):
        """Attempts to build the layer."""
        if obj.built or hasattr(obj.build, "_is_default"):
            obj.built = True
            return True

        if build_input_shape is None:
            build_input_shape = self._infer_inputs(
                node_id, convert_to_shapes=True
            )

        if build_input_shape is not None:
            obj.build(build_input_shape)
            base_layer.Layer.build(obj, build_input_shape)
            return True

        return False

    def get_path(self, node_id):
        return self._node_paths[node_id]

    def finalize_objects(self):
        """Finish setting up Keras objects.

        This function is executed after all objects and functions have been
        created.  Call functions and losses are attached to each layer, and once
        all layers have been fully set up, graph networks are initialized.

        Subclassed models that are revived from the SavedModel are treated like
        layers, and have their call/loss functions attached here.
        """
        # Finish setting up layers and subclassed models. This step attaches
        # call functions and losses to each object, and sets model
        # inputs/outputs.
        layers_revived_from_config = []
        layers_revived_from_saved_model = []
        for node_id, (node, _) in self.loaded_nodes.items():
            if (
                not isinstance(node, base_layer.Layer)
                # Don't finalize models until all layers have finished loading.
                or node_id in self.model_layer_dependencies
            ):
                continue

            self._unblock_model_reconstruction(node_id, node)

            if isinstance(node, input_layer.InputLayer):
                continue
            elif isinstance(node, metrics.Metric):
                continue

            if isinstance(node, (RevivedLayer, RevivedInputLayer)):
                layers_revived_from_saved_model.append(node)
            else:
                layers_revived_from_config.append(node)

        _finalize_saved_model_layers(layers_revived_from_saved_model)
        _finalize_config_layers(layers_revived_from_config)

        # Initialize graph networks, now that layer dependencies have been
        # resolved.
        self._reconstruct_all_models()

    def _unblock_model_reconstruction(self, layer_id, layer):
        """Removes layer from blocking model reconstruction."""
        for model_id, v in self.model_layer_dependencies.items():
            _, layers = v
            if layer_id not in layers:
                continue
            layers[layers.index(layer_id)] = layer
            if all(isinstance(x, base_layer.Layer) for x in layers):
                self._models_to_reconstruct.append(model_id)

    def _reconstruct_all_models(self):
        """Reconstructs the network structure of all models."""
        all_initialized_models = set()
        while self._models_to_reconstruct:
            model_id = self._models_to_reconstruct.pop(0)
            all_initialized_models.add(model_id)
            model, layers = self.model_layer_dependencies[model_id]
            self._reconstruct_model(model_id, model, layers)
            _finalize_config_layers([model])

        if all_initialized_models != set(self.model_layer_dependencies.keys()):
            # This should not happen.
            uninitialized_model_ids = (
                set(self.model_layer_dependencies.keys())
                - all_initialized_models
            )
            uninitialized_model_names = [
                self.model_layer_dependencies[model_id][0].name
                for model_id in uninitialized_model_ids
            ]
            raise ValueError(
                "Error loading model(s) in the SavedModel format. "
                "The following model(s) could not be initialized: "
                f"{uninitialized_model_names}"
            )

    def _reconstruct_model(self, model_id, model, layers):
        """Reconstructs the network structure."""
        config = json_utils.decode(self._metadata[model_id].metadata)["config"]

        # Set up model inputs
        if model.inputs:
            # Inputs may already be created if the model is instantiated in
            # another object's __init__.
            pass
        elif isinstance(model, models_lib.Sequential):
            if not layers or not isinstance(layers[0], input_layer.InputLayer):
                if config["layers"][0]["class_name"] == "InputLayer":
                    layers.insert(
                        0,
                        input_layer.InputLayer.from_config(
                            config["layers"][0]["config"]
                        ),
                    )
                elif "batch_input_shape" in config["layers"][0]["config"]:
                    batch_input_shape = config["layers"][0]["config"][
                        "batch_input_shape"
                    ]
                    layers.insert(
                        0,
                        input_layer.InputLayer(
                            input_shape=batch_input_shape[1:],
                            batch_size=batch_input_shape[0],
                            dtype=layers[0].dtype,
                            name=layers[0].name + "_input",
                        ),
                    )
            model.__init__(layers, name=config["name"])
            if not model.inputs:
                first_layer = self._get_child_layer_node_ids(model_id)[0]
                input_specs = self._infer_inputs(first_layer)
                input_shapes = self._infer_inputs(
                    first_layer, convert_to_shapes=True
                )
                model._set_inputs(input_specs)
                if not model.built and not isinstance(input_specs, dict):
                    model.build(input_shapes)
        else:  # Reconstruct functional model
            (
                inputs,
                outputs,
                created_layers,
            ) = functional_lib.reconstruct_from_config(
                config, created_layers={layer.name: layer for layer in layers}
            )
            model.__init__(inputs, outputs, name=config["name"])
            functional_lib.connect_ancillary_layers(model, created_layers)

        # Set model dtype.
        _set_network_attributes_from_metadata(model)

        # Unblock models that are dependent on this model.
        self._unblock_model_reconstruction(model_id, model)

    def _get_child_layer_node_ids(self, node_id):
        """Returns the node ids of each layer in a Sequential/Functional
        model."""
        # Sequential and Functional track layers with names following the format
        # "layer-N". Use this to generate the list of layers.
        num_layers = 0
        child_layers = {}
        pattern = re.compile("layer-(\\d+)")

        for child in self._proto.nodes[node_id].children:
            m = pattern.match(child.local_name)
            if m is None:
                continue
            layer_n = int(m.group(1))
            num_layers = max(layer_n + 1, num_layers)
            child_layers[layer_n] = child.node_id

        ordered = []
        for n in range(num_layers):
            child = child_layers.get(n)
            if child is None:
                break
            ordered.append(child)
        return ordered

    def _search_for_child_node(self, parent_id, path_to_child):
        """Returns node id of child node.

        A helper method for traversing the object graph proto.

        As an example, say that the object graph proto in the SavedModel
        contains an object with the following child and grandchild attributes:

        `parent.child_a.child_b`

        This method can be used to retrieve the node id of `child_b` using the
        parent's node id by calling:

        `_search_for_child_node(parent_id, ['child_a', 'child_b'])`.

        Args:
          parent_id: node id of parent node
          path_to_child: list of children names.

        Returns:
          node_id of child, or None if child isn't found.
        """
        if not path_to_child:
            return parent_id

        for child in self._proto.nodes[parent_id].children:
            if child.local_name == path_to_child[0]:
                return self._search_for_child_node(
                    child.node_id, path_to_child[1:]
                )
        return None

    def _infer_inputs(self, layer_node_id, convert_to_shapes=False):
        """Infers input shape of layer from SavedModel functions."""
        call_fn_id = self._search_for_child_node(
            layer_node_id, ["call_and_return_all_conditional_losses"]
        )
        if call_fn_id is None:
            return None

        concrete_functions = self._proto.nodes[
            call_fn_id
        ].function.concrete_functions
        if not concrete_functions:
            return None
        call_fn_name = concrete_functions[0]
        call_fn_proto = self._proto.concrete_functions[call_fn_name]
        structured_input_signature = tf.__internal__.saved_model.decode_proto(
            call_fn_proto.canonicalized_input_signature
        )
        inputs = structured_input_signature[0][0]
        if convert_to_shapes:
            return tf.nest.map_structure(lambda spec: spec.shape, inputs)
        else:
            return inputs

    def _config_node_setter(self, setter):
        """Creates edges for nodes that are recreated from config."""

        def setattr_wrapper(obj, name, value):
            # Avoid overwriting attributes of objects recreated from the config.
            if obj._lookup_dependency(name) is None:
                setter(obj, name, value)

        return setattr_wrapper


def _finalize_saved_model_layers(layers):
    """Runs the final steps of loading Keras Layers from SavedModel."""

    # 1. Set up call functions for all layers initialized from the SavedModel (
    # and not the config)
    for layer in layers:
        layer.built = True
        layer_call = getattr(
            _get_keras_attr(layer), "call_and_return_conditional_losses", None
        )
        if layer_call and layer_call.concrete_functions:
            call_spec = layer_utils.CallFunctionSpec(
                tf_inspect.getfullargspec(layer_call)
            )
            layer.call = utils.use_wrapped_call(
                layer, layer_call, call_spec, return_method=True
            )
            expects_training_arg = layer._serialized_attributes["metadata"][
                "expects_training_arg"
            ]
            if "training" in layer_call.function_spec.arg_names:
                # This could change the value of `expects_training_arg` if this
                # layer doesn't expect a training arg, but has a child layer
                # that does.
                expects_training_arg = True
            layer._init_call_fn_args(expects_training_arg)
        else:
            layer.call = types.MethodType(
                _unable_to_call_layer_due_to_serialization_issue, layer
            )

    for layer in layers:
        # 2. Set model inputs and outputs.
        if isinstance(layer, RevivedNetwork):
            _set_network_attributes_from_metadata(layer)

            if hasattr(
                _get_keras_attr(layer), "call_and_return_conditional_losses"
            ):
                call_fn = _get_keras_attr(
                    layer
                ).call_and_return_conditional_losses
                if not call_fn.concrete_functions:
                    continue
                if call_fn.input_signature is None:
                    args, kwargs = infer_inputs_from_restored_call_function(
                        call_fn
                    )
                    args = list(args)
                    inputs = args.pop(0)
                else:
                    args = call_fn.input_signature
                    args = list(args)
                    inputs = args.pop(0)
                    kwargs = None
                layer._set_save_spec(inputs, args, kwargs)

                # V1 models require calling _set_inputs to set the `.inputs`
                # attr.  Skip this step when there are multiple tensor inputs
                # (this behavior is not well supported in V1 models).
                if not any(
                    isinstance(x, tf.TensorSpec)
                    for x in tf.nest.flatten([args, kwargs])
                ):
                    layer._set_inputs(inputs)

        # 3. Add losses that aren't generated by the layer.call function.
        _restore_layer_unconditional_losses(layer)
        _restore_layer_activation_loss(layer)

        # 4. Restore metrics list
        _restore_layer_metrics(layer)


def _unable_to_call_layer_due_to_serialization_issue(
    layer, *unused_args, **unused_kwargs
):
    """Replaces the `layer.call` if the layer was not fully serialized.

    Keras Model/Layer serialization is relatively relaxed because SavedModels
    are not always loaded back as keras models. Thus, when there is an issue
    tracing a non-signature function, a warning is logged instead of raising an
    error. This results in a SavedModel where the model's call function is
    saved, but the internal layer call functions are not.

    When deserialized with `tf.keras.models.load_model`, the internal layers
    which do not have serialized call functions should raise an error when
    called.

    Args:
      layer: Layer without the serialized call function.

    Raises:
      ValueError
    """

    raise ValueError(
        f"Cannot call custom layer {layer.name} of type {type(layer)}, because "
        "the call function was not serialized to the SavedModel."
        "Please try one of the following methods to fix this issue:"
        "\n\n(1) Implement `get_config` and `from_config` in the layer/model "
        "class, and pass the object to the `custom_objects` argument when "
        "loading the model. For more details, see: "
        "https://www.tensorflow.org/guide/keras/save_and_serialize"
        "\n\n(2) Ensure that the subclassed model or layer overwrites `call` "
        "and not `__call__`. The input shape and dtype will be automatically "
        "recorded when the object is called, and used when saving. To manually "
        "specify the input shape/dtype, decorate the call function with "
        "`@tf.function(input_signature=...)`."
    )


def _finalize_config_layers(layers):
    """Runs the final steps of loading Keras Layers from config."""
    for layer in layers:
        # It is assumed that layers define their unconditional losses after
        # being recreated from the config and built. The exceptions to this are
        # Functional and Sequential models, which only store conditional losses
        # (losses dependent on the inputs) in the config. Unconditional losses
        # like weight regularization must be revived from the SavedModel.
        if _is_graph_network(layer):
            _restore_layer_unconditional_losses(layer)

        # Some layers, like Dense, record their activation loss function in the
        # config. However, not all layers do this, so the activation loss may be
        # missing when restored from the config/hdf5.
        # TODO(kathywu): Investigate ways to improve the config to ensure
        # consistent loading behavior between HDF5 and SavedModel.
        _restore_layer_activation_loss(layer)

        # Restore metrics list.
        _restore_layer_metrics(layer)

        # Restore RNN layer states.
        if (
            isinstance(layer, base_rnn.RNN)
            and layer.stateful
            and hasattr(_get_keras_attr(layer), "states")
        ):
            layer.states = getattr(_get_keras_attr(layer), "states", None)
            for variable in tf.nest.flatten(layer.states):
                backend.track_variable(variable)

        # Perform any layer defined finalization of the layer state.
        layer.finalize_state()


def _finalize_metric(metric):
    metric.update_state = types.MethodType(
        metrics_utils.update_state_wrapper(metric.keras_api.update_state),
        metric,
    )
    metric.result = metric.keras_api.result


def _restore_layer_unconditional_losses(layer):
    """Restore unconditional losses from SavedModel."""
    if hasattr(_get_keras_attr(layer), "layer_regularization_losses"):
        losses = getattr(
            _get_keras_attr(layer), "layer_regularization_losses", []
        )
    else:
        # Some earlier SavedModels may not have layer_regularization_losses
        # serialized separately. Fall back to using the regularization_losses
        # list if it does not exist.
        losses = layer._serialized_attributes.get("regularization_losses", [])
    for loss in losses:
        layer.add_loss(loss)


def _restore_layer_activation_loss(layer):
    """Restore actiation loss from SavedModel."""
    # Use wrapped activity regularizer function if the layer's activity
    # regularizer wasn't created during initialization.
    activity_regularizer = getattr(
        _get_keras_attr(layer), "activity_regularizer_fn", None
    )
    if activity_regularizer and not layer.activity_regularizer:
        try:
            layer.activity_regularizer = activity_regularizer
        except AttributeError:
            # This may happen if a layer wrapper is saved with an activity
            # regularizer. The wrapper object's activity regularizer is
            # unsettable.
            pass


def revive_custom_object(identifier, metadata):
    """Revives object from SavedModel."""
    if tf.compat.v1.executing_eagerly_outside_functions():
        model_class = training_lib.Model
    else:
        model_class = training_lib_v1.Model

    revived_classes = {
        constants.INPUT_LAYER_IDENTIFIER: (
            RevivedInputLayer,
            input_layer.InputLayer,
        ),
        constants.LAYER_IDENTIFIER: (RevivedLayer, base_layer.Layer),
        constants.MODEL_IDENTIFIER: (RevivedNetwork, model_class),
        constants.NETWORK_IDENTIFIER: (
            RevivedNetwork,
            functional_lib.Functional,
        ),
        constants.SEQUENTIAL_IDENTIFIER: (
            RevivedNetwork,
            models_lib.Sequential,
        ),
    }
    parent_classes = revived_classes.get(identifier, None)

    class_name = tf.compat.as_str(metadata["class_name"])
    if parent_classes is not None:
        parent_classes = revived_classes[identifier]
        revived_cls = type(class_name, parent_classes, {})
        return revived_cls._init_from_metadata(metadata)
    else:
        raise ValueError(
            f'Unable to restore custom object of class "{class_name}" '
            f"(type {identifier}). Please make sure that this class is "
            "included in the `custom_objects` arg when calling `load_model()`. "
            "Also, check that the class implements `get_config` and "
            f"`from_config`.\n\nComplete metadata: {metadata}"
        )


def _restore_layer_metrics(layer):
    metrics_list = getattr(_get_keras_attr(layer), "layer_metrics", {})
    layer_metrics = {m.name: m for m in layer._metrics}
    for name, metric in metrics_list.items():
        if name not in layer_metrics:
            # Metrics may be added during initialization/building of custom
            # layers.
            layer._metrics.append(metric)


# TODO(kathywu): Centrally define keys and functions for both  serialization and
# deserialization.
class RevivedLayer:
    """Keras layer loaded from a SavedModel."""

    @classmethod
    def _init_from_metadata(cls, metadata):
        """Create revived layer from metadata stored in the SavedModel proto."""
        init_args = dict(name=metadata["name"], trainable=metadata["trainable"])
        if metadata.get("dtype") is not None:
            init_args["dtype"] = metadata["dtype"]
        if metadata.get("batch_input_shape") is not None:
            init_args["batch_input_shape"] = metadata["batch_input_shape"]

        revived_obj = cls(**init_args)

        with utils.no_automatic_dependency_tracking_scope(revived_obj):

            revived_obj._call_spec.expects_training_arg = metadata[
                "expects_training_arg"
            ]
            config = metadata.get("config")
            if serialization.validate_config(config):
                revived_obj._config = config
            if metadata.get("input_spec") is not None:
                revived_obj.input_spec = recursively_deserialize_keras_object(
                    metadata["input_spec"],
                    module_objects={"InputSpec": input_spec.InputSpec},
                )
            if metadata.get("activity_regularizer") is not None:
                revived_obj.activity_regularizer = regularizers.deserialize(
                    metadata["activity_regularizer"]
                )
            if metadata.get("_is_feature_layer") is not None:
                revived_obj._is_feature_layer = metadata["_is_feature_layer"]
            if metadata.get("stateful") is not None:
                revived_obj.stateful = metadata["stateful"]
            if metadata.get("autocast") is not None:
                revived_obj._autocast = metadata["autocast"]
            if metadata.get("preserve_input_structure_in_config") is not None:
                revived_obj._preserve_input_structure_in_config = metadata[
                    "preserve_input_structure_in_config"
                ]

        return revived_obj, _revive_setter

    @property
    def keras_api(self):
        return self._serialized_attributes.get(constants.KERAS_ATTR, None)

    def get_config(self):
        if hasattr(self, "_config"):
            return self._config
        else:
            raise NotImplementedError


def _revive_setter(layer, name, value):
    """Setter function that saves some attributes to separate dictionary."""
    # Many attributes in the SavedModel conflict with properties defined in
    # Layer and Model. Save these attributes to a separate dictionary.
    if name in PUBLIC_ATTRIBUTES:

        if isinstance(value, tf.__internal__.tracking.Trackable):
            layer._track_trackable(value, name=name)
        layer._serialized_attributes[name] = value

    elif (
        isinstance(layer, functional_lib.Functional)
        and re.match(r"^layer(_with_weights)?-[\d+]", name) is not None
    ):
        # Edges named "layer-n" or "layer_with_weights-n", which are tracked in
        # network._track_layers, should not be added as an attribute. They
        # should be temporarily added as a dependency so that checkpointed
        # values can be restored. These dependencies are manually deleted in
        # KerasObjectLoader.del_tracking.

        # Set `overwrite=True` in the case that `layer` already tracks a
        # different layer-n. This may cause variable values to not be loaded
        # properly in the original layer-n, but we already warn the users about
        # this (ctrl-f "shared between different layers/models").
        layer._track_trackable(value, name, overwrite=True)
    elif getattr(layer, name, None) is not None:
        # Don't overwrite already defined attributes.
        pass
    else:
        setattr(layer, name, value)


class RevivedInputLayer:
    """InputLayer loaded from a SavedModel."""

    @classmethod
    def _init_from_metadata(cls, metadata):
        """Revives the saved InputLayer from the Metadata."""
        init_args = dict(
            name=metadata["name"],
            dtype=metadata["dtype"],
            sparse=metadata["sparse"],
            ragged=metadata["ragged"],
            batch_input_shape=metadata["batch_input_shape"],
        )
        revived_obj = cls(**init_args)
        with utils.no_automatic_dependency_tracking_scope(revived_obj):
            revived_obj._config = metadata["config"]

        return revived_obj, setattr

    def get_config(self):
        return self._config


def recursively_deserialize_keras_object(config, module_objects=None):
    """Deserialize Keras object from a nested structure."""
    if isinstance(config, dict):
        if "class_name" in config:
            return serialization.deserialize_keras_object(
                config, module_objects=module_objects
            )
        else:
            return {
                key: recursively_deserialize_keras_object(
                    config[key], module_objects
                )
                for key in config
            }
    elif isinstance(config, (tuple, list)):
        return [
            recursively_deserialize_keras_object(x, module_objects)
            for x in config
        ]
    else:
        raise ValueError(
            "Unable to decode Keras layer config. Config should be a "
            f"dictionary, tuple or list. Received: config={config}"
        )


def infer_inputs_from_restored_call_function(fn):
    """Returns TypeSpec of inputs from a restored call function.

    Args:
      fn: Restored layer call function. It is assumed that `fn` has at least one
        concrete function and that the inputs are in the first argument.

    Returns:
      TypeSpec of call function inputs in the form of (args, kwargs)
    """

    def common_spec(x, y):
        if not isinstance(x, tf.TypeSpec):
            # Doesn't particularly matter what is returned in this case because
            # the result will be filtered out in _set_input_shape.
            return x

        result = x._without_tensor_names().most_specific_common_supertype(
            [y._without_tensor_names()]
        )
        if result is None:
            # Please file a bug if you are being hindered by this error.
            raise TypeError(f"No common supertype of {x} and {y}.")
        return result

    spec = fn.concrete_functions[0].structured_input_signature
    for concrete in fn.concrete_functions[1:]:
        spec2 = concrete.structured_input_signature
        spec = tf.nest.map_structure(common_spec, spec, spec2)
    return spec


class RevivedNetwork(RevivedLayer):
    """Keras network of layers loaded from a SavedModel."""

    @classmethod
    def _init_from_metadata(cls, metadata):
        """Create revived network from metadata stored in the SavedModel
        proto."""
        revived_obj = cls(name=metadata["name"])

        # Store attributes revived from SerializedAttributes in a un-tracked
        # dictionary. The attributes are the ones listed in CommonEndpoints or
        # "keras_api" for keras-specific attributes.
        with utils.no_automatic_dependency_tracking_scope(revived_obj):

            revived_obj._call_spec.expects_training_arg = metadata[
                "expects_training_arg"
            ]
            config = metadata.get("config")
            if serialization.validate_config(config):
                revived_obj._config = config

            if metadata.get("activity_regularizer") is not None:
                revived_obj.activity_regularizer = regularizers.deserialize(
                    metadata["activity_regularizer"]
                )
            if metadata.get("autocast") is not None:
                revived_obj._autocast = metadata["autocast"]

        return revived_obj, _revive_setter


def _set_network_attributes_from_metadata(revived_obj):
    """Sets attributes recorded in the metadata."""
    with utils.no_automatic_dependency_tracking_scope(revived_obj):

        metadata = revived_obj._serialized_attributes["metadata"]
        if metadata.get("dtype") is not None:
            revived_obj._set_dtype_policy(metadata["dtype"])
        revived_obj._trainable = metadata["trainable"]


def _maybe_add_serialized_attributes(layer, metadata):
    # Store attributes revived from SerializedAttributes in a un-tracked
    # dictionary. The attributes are the ones listed in CommonEndpoints or
    # "keras_api" for keras-specific attributes.
    if not hasattr(layer, "_serialized_attributes"):
        with utils.no_automatic_dependency_tracking_scope(layer):
            layer._serialized_attributes = {"metadata": metadata}


def _get_keras_attr(layer):
    return getattr(layer, "_serialized_attributes", {}).get(
        constants.KERAS_ATTR, None
    )
